import warnings
warnings.filterwarnings("ignore")

from fs.opener import opener, OpenerError, Opener
from fs.errors import FSError
from fs.path import splitext, pathsplit, isdotfile, iswildcard

import re
import sys
import platform
import six
from optparse import OptionParser
from collections import defaultdict


if platform.system() == 'Windows':
    def getTerminalSize():
        try:
            ## {{{ http://code.activestate.com/recipes/440694/ (r3)
            from ctypes import windll, create_string_buffer

            # stdin handle is -10
            # stdout handle is -11
            # stderr handle is -12

            h = windll.kernel32.GetStdHandle(-12)
            csbi = create_string_buffer(22)
            res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)

            if res:
                import struct
                (bufx, bufy, curx, cury, wattr,
                 left, top, right, bottom, maxx, maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
                sizex = right - left + 1
                sizey = bottom - top + 1
            else:
                sizex, sizey = 80, 25  # can't determine actual size - return default values
            return sizex, sizey
        except:
            return 80, 25

else:
    def getTerminalSize():
        def ioctl_GWINSZ(fd):
            try:
                import fcntl, termios, struct, os
                cr = struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
            except:
                return None
            return cr
        cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
        if not cr:
            import os
            try:
                fd = os.open(os.ctermid(), os.O_RDONLY)
                cr = ioctl_GWINSZ(fd)
                os.close(fd)
            except:
                pass
        if cr:
            return int(cr[1]), int(cr[0])
        try:
            h, w = os.popen("stty size", "r").read().split()
            return int(w), int(h)
        except:
            pass
        return 80, 25


def _unicode(text):
    if not isinstance(text, unicode):
        return text.decode('ascii', 'replace')
    return text


class Command(object):

    usage = ''
    version = ''

    def __init__(self, usage='', version=''):
        if six.PY3:
            self.output_file = sys.stdout.buffer
            self.error_file = sys.stderr.buffer
        else:
            self.output_file = sys.stdout
            self.error_file = sys.stderr
        self.encoding = getattr(self.output_file, 'encoding', 'utf-8') or 'utf-8'
        self.verbosity_level = 0
        self.terminal_colors = not sys.platform.startswith('win') and self.is_terminal()
        if self.is_terminal():
            w, _h = getTerminalSize()
            self.terminal_width = w
        else:
            self.terminal_width = 80
        self.name = self.__class__.__name__.lower()

    def is_terminal(self):
        try:
            return self.output_file.isatty()
        except AttributeError:
            return False

    def wrap_dirname(self, dirname):
        if not self.terminal_colors:
            return dirname
        return '\x1b[1;34m%s\x1b[0m' % dirname

    def wrap_error(self, msg):
        if not self.terminal_colors:
            return msg
        return '\x1b[31m%s\x1b[0m' % msg

    def wrap_filename(self, fname):
        fname = _unicode(fname)
        if not self.terminal_colors:
            return fname
        if '://' in fname:
            return fname
#        if '.' in fname:
#            name, ext = splitext(fname)
#            fname = u'%s\x1b[36m%s\x1b[0m' % (name, ext)
        if isdotfile(fname):
            fname = '\x1b[33m%s\x1b[0m' % fname
        return fname

    def wrap_faded(self, text):
        text = _unicode(text)
        if not self.terminal_colors:
            return text
        return u'\x1b[2m%s\x1b[0m' % text

    def wrap_link(self, text):
        if not self.terminal_colors:
            return text
        return u'\x1b[1;33m%s\x1b[0m' % text

    def wrap_strong(self, text):
        if not self.terminal_colors:
            return text
        return u'\x1b[1m%s\x1b[0m' % text

    def wrap_table_header(self, name):
        if not self.terminal_colors:
            return name
        return '\x1b[1;32m%s\x1b[0m' % name

    def highlight_fsurls(self, text):
        if not self.terminal_colors:
            return text
        re_fs = r'(\S*?://\S*)'

        def repl(matchobj):
            fs_url = matchobj.group(0)
            return self.wrap_link(fs_url)
        return re.sub(re_fs, repl, text)

    def open_fs(self, fs_url, writeable=False, create_dir=False):
        fs, path = opener.parse(fs_url, writeable=writeable, create_dir=create_dir)
        fs.cache_hint(True)
        return fs, path

    def expand_wildcard(self, fs, path):
        if path is None:
            return [], []
        pathname, resourcename = pathsplit(path)
        if iswildcard(resourcename):
            dir_paths = fs.listdir(pathname,
                                   wildcard=resourcename,
                                   absolute=True,
                                   dirs_only=True)

            file_paths = fs.listdir(pathname,
                                    wildcard=resourcename,
                                    absolute=True,
                                    files_only=True)
            return dir_paths, file_paths

        else:
            if fs.isdir(path):
                #file_paths = fs.listdir(path,
                #                        absolute=True)
                return [path], []
            return [], [path]

    def get_resources(self, fs_urls, dirs_only=False, files_only=False, single=False):

        fs_paths = [self.open_fs(fs_url) for fs_url in fs_urls]
        resources = []

        for fs, path in fs_paths:
            if path and iswildcard(path):
                if not files_only:
                    dir_paths = fs.listdir(wildcard=path, dirs_only=True)
                    for path in dir_paths:
                        resources.append([fs, path, True])
                if not dirs_only:
                    file_paths = fs.listdir(wildcard=path, files_only=True)
                    for path in file_paths:
                        resources.append([fs, path, False])
            else:
                path = path or '/'
                is_dir = fs.isdir(path)
                resource = [fs, path, is_dir]
                if not files_only and not dirs_only:
                    resources.append(resource)
                elif files_only and not is_dir:
                    resources.append(resource)
                elif dirs_only and is_dir:
                    resources.append(resource)

            if single:
                break

        return resources

    def ask(self, msg):
        return raw_input('%s: %s ' % (self.name, msg))

    def text_encode(self, text):
        if not isinstance(text, unicode):
            text = text.decode('ascii', 'replace')
        text = text.encode(self.encoding, 'replace')
        return text

    def output(self, msgs, verbose=False):
        if verbose and not self.options.verbose:
            return
        if isinstance(msgs, basestring):
            msgs = (msgs,)
        for msg in msgs:
            self.output_file.write(self.text_encode(msg))

    def output_table(self, table, col_process=None, verbose=False):
        if verbose and not self.verbose:
            return
        if col_process is None:
            col_process = {}

        max_row_widths = defaultdict(int)

        for row in table:
            for col_no, col in enumerate(row):
                max_row_widths[col_no] = max(max_row_widths[col_no], len(col))

        lines = []
        for row in table:
            out_col = []
            for col_no, col in enumerate(row):
                td = col.ljust(max_row_widths[col_no])
                if col_no in col_process:
                    td = col_process[col_no](td)
                out_col.append(td)
            lines.append(self.text_encode('%s\n' % '  '.join(out_col).rstrip()))
        for l in lines:
            self.output_file.write(l)
        #self.output(''.join(lines))

    def error(self, *msgs):
        for msg in msgs:
            self.error_file.write("{}: {}".format(self.name, msg).encode(self.encoding))

    def get_optparse(self):
        optparse = OptionParser(usage=self.usage, version=self.version)
        optparse.add_option('--debug', dest='debug', action="store_true", default=False,
                            help="Show debug information", metavar="DEBUG")
        optparse.add_option('-v', '--verbose', dest='verbose', action="store_true", default=False,
                            help="make output verbose", metavar="VERBOSE")
        optparse.add_option('--listopeners', dest='listopeners', action="store_true", default=False,
                            help="list all FS openers", metavar="LISTOPENERS")
        optparse.add_option('--fs', dest='fs', action='append', type="string",
                            help="import an FS opener e.g --fs foo.bar.MyOpener", metavar="OPENER")
        return optparse

    def list_openers(self):

        opener_table = []

        for fs_opener in opener.openers.itervalues():
            names = fs_opener.names
            desc = getattr(fs_opener, 'desc', '')
            opener_table.append((names, desc))

        opener_table.sort(key=lambda r: r[0])

        def wrap_line(text):

            lines = text.split('\n')
            for line in lines:
                words = []
                line_len = 0
                for word in line.split():
                    if word == '*':
                        word = ' *'
                    if line_len + len(word) > self.terminal_width:
                        self.output((self.highlight_fsurls(' '.join(words)), '\n'))
                        del words[:]
                        line_len = 0
                    words.append(word)
                    line_len += len(word) + 1
                if words:
                    self.output(self.highlight_fsurls(' '.join(words)))
                self.output('\n')

        for names, desc in opener_table:
            self.output(('-' * self.terminal_width, '\n'))
            proto = ', '.join([n + '://' for n in names])
            self.output((self.wrap_dirname('[%s]' % proto), '\n\n'))
            if not desc.strip():
                desc = "No information available"
            wrap_line(desc)
            self.output('\n')

    def run(self):
        parser = self.get_optparse()
        options, args = parser.parse_args()
        self.options = options

        if options.listopeners:
            self.list_openers()
            return 0

        ilocals = {}
        if options.fs:
            for import_opener in options.fs:
                module_name, opener_class = import_opener.rsplit('.', 1)
                try:
                    opener_module = __import__(module_name, globals(), ilocals, [opener_class], -1)
                except ImportError:
                    self.error("Unable to import opener %s\n" % import_opener)
                    return 0

                new_opener = getattr(opener_module, opener_class)

                try:
                    if not issubclass(new_opener, Opener):
                        self.error('%s is not an fs.opener.Opener\n' % import_opener)
                        return 0
                except TypeError:
                    self.error('%s is not an opener class\n' % import_opener)
                    return 0

                if options.verbose:
                    self.output('Imported opener %s\n' % import_opener)

                opener.add(new_opener)

        if not six.PY3:
            args = [unicode(arg, sys.getfilesystemencoding()) for arg in args]
        self.verbose = options.verbose
        try:
            return self.do_run(options, args) or 0
        except FSError, e:
            self.error(self.wrap_error(unicode(e)) + '\n')
            if options.debug:
                raise
            return 1
        except KeyboardInterrupt:
            if self.is_terminal():
                self.output("\n")
            return 0
        except SystemExit:
            return 0
        except Exception, e:
            self.error(self.wrap_error('Error - %s\n' % unicode(e)))
            if options.debug:
                raise
            return 1


if __name__ == "__main__":
    command = Command()
    sys.exit(command.run())


'''
    @author: Marek Palatinus <marek@palatinus.cz>
    @license: Public domain
    
    DebugFS is a wrapper around filesystems to help developers
    debug their work. I wrote this class mainly for debugging
    TahoeLAFS and for fine tuning TahoeLAFS over Dokan with higher-level
    aplications like Total Comander, Winamp etc. Did you know
    that Total Commander need to open file before it delete them? :-)
    
    I hope DebugFS can be helpful also for other filesystem developers,
    especially for those who are trying to implement their first one (like me).
    
    DebugFS prints to stdout (by default) all attempts to
    filesystem interface, prints parameters and results.
    
    Basic usage:
        fs = DebugFS(OSFS('~'), identifier='OSFS@home', \
                skip=('_lock', 'listdir', 'listdirinfo'))
        print fs.listdir('.')
        print fs.unsupportedfunction()
    
    Error levels:
        DEBUG: Print everything (asking for methods, calls, response, exception)
        INFO: Print calls, responses, exception
        ERROR: Print only exceptions 
        CRITICAL: Print only exceptions not derived from fs.errors.FSError
        
    How to change error level:
        import logging
        logger = logging.getLogger('fs.debugfs')
        logger.setLevel(logging.CRITICAL)
        fs = DebugFS(OSFS('~')
        print fs.listdir('.')
    
'''
import logging
from logging import DEBUG, INFO, ERROR, CRITICAL
import sys

import fs
from fs.errors import FSError

logger = fs.getLogger('fs.debugfs')
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler())

class DebugFS(object):
    def __init__(self, fs, identifier=None, skip=(), verbose=True):
        '''
            fs - Reference to object to debug
            identifier - Custom string-like object will be added
                to each log line as identifier.
            skip - list of method names which DebugFS should not log
        '''        
        self.__wrapped_fs = fs
        self.__identifier = identifier
        self.__skip = skip
        self.__verbose = verbose 
        super(DebugFS, self).__init__()
        
    def __log(self, level, message):
        if self.__identifier:
            logger.log(level, '(%s) %s' % (self.__identifier, message))
        else:            
            logger.log(level, message)
        
    def __parse_param(self, value):
        if isinstance(value, basestring):
            if len(value) > 60:
                value = "%s ... (length %d)" % (repr(value[:60]), len(value))
            else:
                value = repr(value) 
        elif isinstance(value, list):
            value = "%s (%d items)" % (repr(value[:3]), len(value))
        elif isinstance(value, dict):
            items = {}
            for k, v in value.items()[:3]:
                items[k] = v
            value = "%s (%d items)" % (repr(items), len(value))
        else:
            value = repr(value)
        return value
    
    def __parse_args(self, *arguments, **kwargs):
        args = [self.__parse_param(a) for a in arguments]
        for k, v in kwargs.items():
            args.append("%s=%s" % (k, self.__parse_param(v)))
        
        args = ','.join(args)
        if args: args = "(%s)" % args
        return args
        
    def __report(self, msg, key, value, *arguments, **kwargs):
        if key in self.__skip: return
        args = self.__parse_args(*arguments, **kwargs)
        value = self.__parse_param(value)
        self.__log(INFO, "%s %s%s -> %s" % (msg, str(key), args, value))
    
    def __getattr__(self, key):
        
        if key.startswith('__'):
            # Internal calls, nothing interesting
            return object.__getattribute__(self, key)

        try:
            attr = getattr(self.__wrapped_fs, key)
        except AttributeError, e:
            self.__log(DEBUG, "Asking for not implemented method %s" % key)
            raise e
        except Exception, e:
            self.__log(CRITICAL, "Exception %s: %s" % \
                     (e.__class__.__name__, str(e)))
            raise e
                
        if not callable(attr):
            if key not in self.__skip:
                self.__report("Get attribute", key, attr)
            return attr
        
        def _method(*args, **kwargs):
            try:
                value = attr(*args, **kwargs)                
                self.__report("Call method", key, value, *args, **kwargs)
            except FSError, e:
                self.__log(ERROR, "Call method %s%s -> Exception %s: %s" % \
                             (key, self.__parse_args(*args, **kwargs), \
                             e.__class__.__name__, str(e)))
                (exc_type,exc_inst,tb) = sys.exc_info()
                raise e, None, tb
            except Exception, e:
                self.__log(CRITICAL,
                         "Call method %s%s -> Non-FS exception %s: %s" %\
                         (key, self.__parse_args(*args, **kwargs), \
                         e.__class__.__name__, str(e)))
                (exc_type,exc_inst,tb) = sys.exc_info()
                raise e, None, tb
            return value
        
        if self.__verbose:
            if key not in self.__skip:
                self.__log(DEBUG, "Asking for method %s" % key)
        return _method



#------------------------------------------------------------------------------
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   01/24/2002
#
#------------------------------------------------------------------------------

""" Dynamically construct wxPython Menus or MenuBars from a supplied string
    description of the menu.

Menu Description Syntax::

    submenu_label {help_string}
        menuitem_label | accelerator {help_string} [~/-name]: code

*submenu_label*
    Label of a sub menu
*menuitem_label*
    Label of a menu item
{*help_string*}
    Help string to display on the status line (optional)
*accelerator*
    Accelerator key (e.g., Ctrl-C) (The '|' and keyname are optional, but must
    be used together.)
[~]
    The menu item is checkable, but is not checked initially (optional)
[/]
    The menu item is checkable, and is checked initially (optional)
[-]
    The menu item disabled initially (optional)
[*name*]
    Symbolic name used to refer to menu item (optional)
*code*
    Python code invoked when menu item is selected

A line beginning with a hyphen (-) is interpreted as a menu separator.
"""

#=========================================================================
#  Imports:
#=========================================================================

import wx
import re
import string

#=========================================================================
#  Constants:
#=========================================================================

help_pat = re.compile(r'(.*){(.*)}(.*)')
options_pat = re.compile(r'(.*)\[(.*)\](.*)')

# Mapping of key name strings to wxPython key codes
key_map = {
    'F1': wx.WXK_F1,
    'F2': wx.WXK_F2,
    'F3': wx.WXK_F3,
    'F4': wx.WXK_F4,
    'F5': wx.WXK_F5,
    'F6': wx.WXK_F6,
    'F7': wx.WXK_F7,
    'F8': wx.WXK_F8,
    'F9': wx.WXK_F9,
    'F10': wx.WXK_F10,
    'F11': wx.WXK_F11,
    'F12': wx.WXK_F12
}

#-------------------------------------------------------------------------
#  'MakeMenu' class:
#-------------------------------------------------------------------------


class MakeMenu:
    """ Manages creation of menus.
    """
    # Initialize the globally unique menu ID:
    cur_id = 1000

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, desc, owner, popup=False, window=None):
        """ Initializes the object.
        """
        self.owner = owner
        if window is None:
            window = owner
        self.window = window
        self.indirect = getattr(owner, 'call_menu', None)
        self.names = {}
        self.desc = desc.split('\n')
        self.index = 0
        self.keys = []
        if popup:
            self.menu = menu = wx.Menu()
            self.parse(menu, -1)
        else:
            self.menu = menu = wx.MenuBar()
            self.parse(menu, -1)
            window.SetMenuBar(menu)
            if len(self.keys) > 0:
                window.SetAcceleratorTable(wx.AcceleratorTable(self.keys))

    #-------------------------------------------------------------------------
    #  Recursively parses menu items from the description:
    #-------------------------------------------------------------------------

    def parse(self, menu, indent):
        """ Recursively parses menu items from the description.
        """

        while True:

            # Make sure we have not reached the end of the menu description
            # yet:
            if self.index >= len(self.desc):
                return

            # Get the next menu description line and check its indentation:
            dline = self.desc[self.index]
            line = dline.lstrip()
            indented = len(dline) - len(line)
            if indented <= indent:
                return

            # Indicate that the current line has been processed:
            self.index += 1

            # Check for a blank or comment line:
            if (line == '') or (line[0:1] == '#'):
                continue

            # Check for a menu separator:
            if line[0:1] == '-':
                menu.AppendSeparator()
                continue

            # Allocate a new menu ID:
            MakeMenu.cur_id += 1
            cur_id = MakeMenu.cur_id

            # Extract the help string (if any):
            help = ''
            match = help_pat.search(line)
            if match:
                help = ' ' + match.group(2).strip()
                line = match.group(1) + match.group(3)

            # Check for a menu item:
            col = line.find(':')
            if col >= 0:
                handler = line[col + 1:].strip()
                if handler != '':
                    if self.indirect:
                        self.indirect(cur_id, handler)
                        handler = self.indirect
                    else:
                        try:
                            exec (
                                'def handler(event,self=self.owner):\n %s\n' %
                                handler)
                        except:
                            handler = null_handler
                else:
                    try:
                        exec 'def handler(event,self=self.owner):\n%s\n' % (
                            self.get_body(indented), ) in globals()
                    except:
                        handler = null_handler
                wx.EVT_MENU(self.window, cur_id, handler)
                not_checked = checked = disabled = False
                line = line[: col]
                match = options_pat.search(line)
                if match:
                    line = match.group(1) + match.group(3)
                    not_checked, checked, disabled, name = option_check(
                        '~/-', match.group(2).strip())
                    if name != '':
                        self.names[name] = cur_id
                        setattr(self.owner, name, MakeMenuItem(self, cur_id))
                label = line.strip()
                col = label.find('|')
                if col >= 0:
                    key = label[col + 1:].strip()
                    label = '%s%s%s' % (label[: col].strip(), '\t', key)
                    key = key.upper()
                    flag = wx.ACCEL_NORMAL
                    col = key.find('-')
                    if col >= 0:
                        flag = {'CTRL': wx.ACCEL_CTRL,
                                'SHIFT': wx.ACCEL_SHIFT,
                                'ALT': wx.ACCEL_ALT
                                }.get(key[: col].strip(), wx.ACCEL_CTRL)
                        key = key[col + 1:].strip()
                    code = key_map.get(key, None)
                    try:
                        if code is None:
                            code = ord(key)
                        self.keys.append(
                            wx.AcceleratorEntry(flag, code, cur_id))
                    except:
                        pass
                menu.Append(cur_id, label, help, not_checked or checked)
                if checked:
                    menu.Check(cur_id, True)
                if disabled:
                    menu.Enable(cur_id, False)
                continue

            # Else must be the start of a sub menu:
            submenu = wx.Menu()
            label = line.strip()

            # Recursively parse the sub-menu:
            self.parse(submenu, indented)

            # Add the menu to its parent:
            try:
                menu.AppendMenu(cur_id, label, submenu, help)
            except:
                # Handle the case where 'menu' is really a 'MenuBar' (which does
                # not understand 'MenuAppend'):
                menu.Append(submenu, label)

    #-------------------------------------------------------------------------
    #  Returns the body of an inline method:
    #-------------------------------------------------------------------------

    def get_body(self, indent):
        """ Returns the body of an inline method.
        """
        result = []
        while self.index < len(self.desc):
            line = self.desc[self.index]
            if (len(line) - len(line.lstrip())) <= indent:
                break
            result.append(line)
            self.index += 1
        result = '\n'.join(result).rstrip()
        if result != '':
            return result
        return '  pass'

    #-------------------------------------------------------------------------
    #  Returns the id associated with a specified name:
    #-------------------------------------------------------------------------

    def get_id(self, name):
        """ Returns the ID associated with a specified name.
        """
        if isinstance(name, basestring):
            return self.names[name]
        return name

    #-------------------------------------------------------------------------
    #  Checks (or unchecks) a menu item specified by name:
    #-------------------------------------------------------------------------

    def checked(self, name, check=None):
        """ Checks (or unchecks) a menu item specified by name.
        """
        if check is None:
            return self.menu.IsChecked(self.get_id(name))
        self.menu.Check(self.get_id(name), check)

    #-------------------------------------------------------------------------
    #  Enables (or disables) a menu item specified by name:
    #-------------------------------------------------------------------------

    def enabled(self, name, enable=None):
        """ Enables (or disables) a menu item specified by name.
        """
        if enable is None:
            return self.menu.IsEnabled(self.get_id(name))
        self.menu.Enable(self.get_id(name), enable)

    #-------------------------------------------------------------------------
    #  Gets/Sets the label for a menu item:
    #-------------------------------------------------------------------------

    def label(self, name, label=None):
        """ Gets or sets the label for a menu item.
        """
        if label is None:
            return self.menu.GetLabel(self.get_id(name))
        self.menu.SetLabel(self.get_id(name), label)

#-------------------------------------------------------------------------
#  'MakeMenuItem' class:
#-------------------------------------------------------------------------


class MakeMenuItem:
    """ A menu item for a menu managed by MakeMenu.
    """

    def __init__(self, menu, id):
        self.menu = menu
        self.id = id

    def checked(self, check=None):
        return self.menu.checked(self.id, check)

    def toggle(self):
        checked = not self.checked()
        self.checked(checked)
        return checked

    def enabled(self, enable=None):
        return self.menu.enabled(self.id, enable)

    def label(self, label=None):
        return self.menu.label(self.id, label)

#-------------------------------------------------------------------------
#  Determine whether a string contains any specified option characters, and
#  remove them if it does:
#-------------------------------------------------------------------------


def option_check(test, string):
    """ Determines whether a string contains any specified option characters,
    and removes them if it does.
    """
    result = []
    for char in test:
        col = string.find(char)
        result.append(col >= 0)
        if col >= 0:
            string = string[: col] + string[col + 1:]
    return result + [string.strip()]

#-------------------------------------------------------------------------
#  Null menu option selection handler:
#-------------------------------------------------------------------------


def null_handler(event):
    print 'null_handler invoked'



"""
fs.wrapfs.readonlyfs
====================

An FS wrapper class for blocking operations that would modify the FS.

"""

from fs.base import NoDefaultMeta
from fs.wrapfs import WrapFS
from fs.errors import UnsupportedError, NoSysPathError


class ReadOnlyFS(WrapFS):
    """ Makes a FS object read only. Any operation that could potentially modify
    the underlying file system will throw an UnsupportedError

    Note that this isn't a secure sandbox, untrusted code could work around the
    read-only restrictions by getting the base class. Its main purpose is to
    provide a degree of safety if you want to protect an FS object from
    accidental modification.

    """

    def getmeta(self, meta_name, default=NoDefaultMeta):
        if meta_name == 'read_only':
            return True
        return self.wrapped_fs.getmeta(meta_name, default)

    def hasmeta(self, meta_name):
        if meta_name == 'read_only':
            return True
        return self.wrapped_fs.hasmeta(meta_name)

    def getsyspath(self, path, allow_none=False):
        """ Doesn't technically modify the filesystem but could be used to work
        around read-only restrictions. """
        if allow_none:
            return None
        raise NoSysPathError(path)

    def open(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):
        """ Only permit read access """
        if 'w' in mode or 'a' in mode or '+' in mode:
            raise UnsupportedError('write')
        return super(ReadOnlyFS, self).open(path,
                                            mode=mode,
                                            buffering=buffering,
                                            encoding=encoding,
                                            errors=errors,
                                            newline=newline,
                                            line_buffering=line_buffering,
                                            **kwargs)

    def _no_can_do(self, *args, **kwargs):
        """ Replacement method for methods that can modify the file system """
        raise UnsupportedError('write')

    move = _no_can_do
    movedir = _no_can_do
    copy = _no_can_do
    copydir = _no_can_do
    makedir = _no_can_do
    rename = _no_can_do
    setxattr = _no_can_do
    delxattr = _no_can_do
    remove = _no_can_do
    removedir = _no_can_do
    settimes = _no_can_do
    setcontents = _no_can_do
    createfile = _no_can_do



import OpenGL.GL as gl
import math

triangles_drawn = 0
def clear_stats():
    global triangles_drawn
    triangles_drawn = 0

def add_triangle_stats(tris):
    global triangles_drawn
    triangles_drawn += tris

def draw_triangle_strip(vertices, color):
    if color:
        gl.glColor4ub(*color)
    n_vertices = len(vertices)/2
    add_triangle_stats(n_vertices-2)
    gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
    gl.glVertexPointer(2, gl.GL_FLOAT, 0, vertices)
    gl.glDrawArrays(gl.GL_TRIANGLE_STRIP, 0, len(vertices) / 2)
    gl.glDisableClientState(gl.GL_VERTEX_ARRAY)


def draw_round_cap(center, radius, angle):
    n_vertices = 1
    v = [center.x, center.y]

    for theta in xrange(-90, 91, 10):
        at = theta*(math.pi/180) + angle
        x = math.cos(at) * radius + center.x
        y = math.sin(at) * radius + center.y
        n_vertices += 1
        v.append(x)
        v.append(y)

    gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
    gl.glVertexPointer(2, gl.GL_FLOAT, 0, v)
    gl.glDrawArrays(gl.GL_TRIANGLE_FAN, 0, len(v) / 2)
    gl.glDisableClientState(gl.GL_VERTEX_ARRAY)

    add_triangle_stats(n_vertices-1)


def draw_colored_triangles(tris, colors):
    add_triangle_stats(len(tris)/6)
    gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
    gl.glEnableClientState(gl.GL_COLOR_ARRAY)
    gl.glColorPointer(4, gl.GL_UNSIGNED_BYTE, 0, colors)
    gl.glVertexPointer(2, gl.GL_FLOAT, 0, tris)
    gl.glDrawArrays(gl.GL_TRIANGLES, 0, len(tris) / 2)
    gl.glDisableClientState(gl.GL_VERTEX_ARRAY)
    gl.glDisableClientState(gl.GL_COLOR_ARRAY)


def draw_textured_triangles(tris, tex_coords):
    add_triangle_stats(len(tris)/6)
    gl.glColor4f(1, 1, 1, 1)
    gl.glEnable(gl.GL_TEXTURE_2D)
    gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
    gl.glEnableClientState(gl.GL_TEXTURE_COORD_ARRAY)

    gl.glVertexPointer(2, gl.GL_FLOAT, 0, tris)
    gl.glTexCoordPointer(2, gl.GL_FLOAT, 0, tex_coords)
    gl.glDrawArrays(gl.GL_TRIANGLES, 0, len(tris) / 2)
    gl.glDisableClientState(gl.GL_VERTEX_ARRAY)
    gl.glDisableClientState(gl.GL_TEXTURE_COORD_ARRAY)
    gl.glDisable(gl.GL_TEXTURE_2D)


def draw_quad(x, y, w, h):

    points = [x, y,
              x + w, y,
              x + w, y + h,
              x, y + h]

    tex_coords = [0, 1,
                  1, 1,
                  1, 0,
                  0, 0]
    add_triangle_stats(2)
    gl.glColor4f(1, 1, 1, 1)
    gl.glEnable(gl.GL_TEXTURE_2D)
    gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
    gl.glEnableClientState(gl.GL_TEXTURE_COORD_ARRAY)

    gl.glVertexPointer(2, gl.GL_FLOAT, 0, points)
    gl.glTexCoordPointer(2, gl.GL_FLOAT, 0, tex_coords)
    gl.glDrawArrays(gl.GL_QUADS, 0, len(points) / 2)
    gl.glDisableClientState(gl.GL_VERTEX_ARRAY)
    gl.glDisableClientState(gl.GL_TEXTURE_COORD_ARRAY)
    gl.glDisable(gl.GL_TEXTURE_2D)


#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The representation of a name-to-object binding in a context. """


# Enthought libary imports.
from traits.api import Any, HasTraits, Property, Str


class Binding(HasTraits):
    """ The representation of a name-to-object binding in a context. """

    #### 'Binding' interface ##################################################

    # The class name of the object bound to the name in the binding.
    class_name = Property(Str)

    # The name.
    name = Str

    # The object bound to the name in the binding.
    obj = Any

    #### Experimental 'Binding' interface #####################################

    # fixme: These is not part of the JNDI spec, but they do seem startlingly
    # useful!
    #
    # fixme: And, errr, startlingly broken! If the context that the binding
    # is in is required then just look up the name minus the last component!
    #
    # The context that the binding came from.
    context = Any

    # The name of the bound object within the namespace.
    namespace_name = Property(Str)

    #### 'Private' interface ##################################################

    # Shadow trait for the 'class_name' property.
    _class_name = Str

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __str__(self):
        """ Returns an informal string representation of the object. """

        return super(Binding, self).__str__() + '(name=%s, obj=%s)' % (
               self.name, self.obj)

    ###########################################################################
    # 'Binding' interface.
    ###########################################################################

    #### Properties ###########################################################

    # class_name
    def _get_class_name(self):
        """ Returns the class name of the object. """

        if len(self._class_name) == 0:
            if self.obj is None:
                class_name = None

            else:
                klass = self.obj.__class__

                class_name = '%s.%s' % (klass.__module__, klass.__name__)

        return class_name

    def _set_class_name(self, class_name):
        """ Sets the class name of the object. """

        self._class_name = class_name

        return

    # namespace_name
    def _get_namespace_name(self):
        """ Returns the name of the context within its own namespace. """

        if self.context is not None:
            base = self.context.namespace_name

        else:
            base = ''

        if len(base) > 0:
            namespace_name = base + '/' + self.name

        else:
            namespace_name = self.name

        return namespace_name

#### EOF ######################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2012, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Author: Pietro Berkes
#  Date:   Feb 2012
#
#------------------------------------------------------------------------------

"""
Test cases for the UI object.
"""

from traits.has_traits import HasTraits
from traits.trait_types import Str, Int
import traitsui
from traitsui.item import Item
from traitsui.view import View

from traitsui.tests._tools import *


class FooDialog(HasTraits):
    """Test dialog that does nothing useful."""

    my_int = Int(2)
    my_str = Str('hallo')

    traits_view = View(
        Item('my_int'),
        Item('my_str'),
        buttons=['OK']
    )


@skip_if_not_wx
def test_reset_with_destroy_wx():
    # Characterization test:
    # UI.reset(destroy=True) destroys all ui children of the top control

    foo = FooDialog()
    ui = foo.edit_traits()

    ui.reset(destroy=True)

    # the top control is still there
    nose.tools.assert_is_not_none(ui.control)
    # but its children are gone
    nose.tools.assert_equal(len(ui.control.GetChildren()), 0)


@skip_if_not_qt4
def test_reset_with_destroy_qt():
    # Characterization test:
    # UI.reset(destroy=True) destroys all ui children of the top control

    from pyface import qt

    foo = FooDialog()
    ui = foo.edit_traits()

    # decorate children's `deleteLater` function to check that it is called
    # on `reset`. check only with the editor parts (only widgets are scheduled,
    # see traitsui.qt4.toolkit.GUIToolkit.destroy_children)
    for c in ui.control.children():
        c.deleteLater = count_calls(c.deleteLater)

    ui.reset(destroy=True)

    # the top control is still there
    nose.tools.assert_is_not_none(ui.control)

    # but its children are scheduled for removal
    for c in ui.control.children():
        if isinstance(c, qt.QtGui.QWidget):
            nose.tools.assert_equal(c.deleteLater._n_calls, 1)


@skip_if_not_wx
def test_reset_without_destroy_wx():
    # Characterization test:
    # UI.reset(destroy=False) destroys all editor controls, but leaves editors
    # and ui children intact

    import wx

    foo = FooDialog()
    ui = foo.edit_traits()

    nose.tools.assert_equal(len(ui._editors), 2)
    nose.tools.assert_is_instance(ui._editors[0],
                                  traitsui.wx.text_editor.SimpleEditor)
    nose.tools.assert_is_instance(ui._editors[0].control,
                                  wx._controls.TextCtrl)

    ui.reset(destroy=False)

    nose.tools.assert_equal(len(ui._editors), 2)
    nose.tools.assert_is_instance(ui._editors[0],
                                  traitsui.wx.text_editor.SimpleEditor)
    nose.tools.assert_is_none(ui._editors[0].control)

    # children are still there: check first text control
    text_ctrl = ui.control.FindWindowByName('text')
    nose.tools.assert_is_not_none(text_ctrl)


@skip_if_not_qt4
def test_reset_without_destroy_qt():
    # Characterization test:
    # UI.reset(destroy=False) destroys all editor controls, but leaves editors
    # and ui children intact

    from pyface import qt

    foo = FooDialog()
    ui = foo.edit_traits()

    nose.tools.assert_equal(len(ui._editors), 2)
    nose.tools.assert_is_instance(ui._editors[0],
                                  traitsui.qt4.text_editor.SimpleEditor)
    nose.tools.assert_is_instance(ui._editors[0].control,
                                  qt.QtGui.QLineEdit)

    ui.reset(destroy=False)

    nose.tools.assert_equal(len(ui._editors), 2)
    nose.tools.assert_is_instance(ui._editors[0],
                                  traitsui.qt4.text_editor.SimpleEditor)
    nose.tools.assert_is_none(ui._editors[0].control)

    # children are still there: check first text control
    text_ctrl = ui.control.findChild(qt.QtGui.QLineEdit)
    nose.tools.assert_is_not_none(text_ctrl)


@skip_if_not_wx
def test_destroy_after_ok_wx():
    # Behavior: after pressing 'OK' in a dialog, the method UI.finish is
    # called and the view control and its children are destroyed

    import wx

    foo = FooDialog()
    ui = foo.edit_traits()

    # keep reference to the control to check that it was destroyed
    control = ui.control

    # decorate control's `Destroy` function to check that it is called
    control.Destroy = count_calls(control.Destroy)

    # press the OK button and close the dialog
    okbutton = ui.control.FindWindowByName('button')
    click_event = wx.CommandEvent(wx.wxEVT_COMMAND_BUTTON_CLICKED,
                                  okbutton.GetId())
    okbutton.ProcessEvent(click_event)

    nose.tools.assert_is_none(ui.control)
    nose.tools.assert_equal(control.Destroy._n_calls, 1)


@skip_if_not_qt4
def test_destroy_after_ok_qt():
    # Behavior: after pressing 'OK' in a dialog, the method UI.finish is
    # called and the view control and its children are destroyed

    from pyface import qt

    foo = FooDialog()
    ui = foo.edit_traits()

    # keep reference to the control to check that it was deleted
    control = ui.control

    # decorate control's `deleteLater` function to check that it is called
    control.deleteLater = count_calls(control.deleteLater)

    # press the OK button and close the dialog
    okb = control.findChild(qt.QtGui.QPushButton)
    okb.click()

    nose.tools.assert_is_none(ui.control)
    nose.tools.assert_equal(control.deleteLater._n_calls, 1)



#!/usr/bin/env python

"""
Test code for nc_particles

Not very complete

Designed to be run with pytest

"""
# for py2/3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import os

import pytest
import netCDF4
from post_gnome import nc_particles


def test_init():
    """
    Can the classes be intitialized?
    """
    w = nc_particles.Writer('junk_file.nc')
    del w
    print(os.getcwd())
    nc_particles.Reader('sample.nc')


def test_3_unlimited():
    with pytest.raises(ValueError):
        nc_particles.Writer('junk_file2.nc', nc_version=3)


def test_netcdf3():
    w = nc_particles.Writer('junk_file3.nc', num_timesteps=10, nc_version='3')
    w.close()
    nc = netCDF4.Dataset('junk_file3.nc')
    assert nc.file_format == 'NETCDF3_CLASSIC'


def test_netcdf4():
    w = nc_particles.Writer('junk_file4.nc', num_timesteps=10, nc_version=4)
    w.close()
    nc = netCDF4.Dataset('junk_file4.nc')
    assert nc.file_format == 'NETCDF4'


def test_netcdf_wrong():
    with pytest.raises(ValueError):
        nc_particles.Writer('junk_file.nc', nc_version='nc4')


def test_netcdf_wrong_num():
    with pytest.raises(ValueError):
        nc_particles.Writer('junk_file.nc', nc_version='5')


def test_multi_close():
    w = nc_particles.Writer('junk_file5.nc',
                            nc_version=4)
    w.close()
    w.close()



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   06/26/2007
#
#-------------------------------------------------------------------------

""" Traits UI themed button editor.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api \
    import Instance, Str, on_trait_change

from traitsui.api \
    import Theme

from traitsui.ui_traits \
    import ATheme, AView, Image, Position, Spacing

from traitsui.wx.editor \
    import Editor

from traitsui.basic_editor_factory \
    import BasicEditorFactory

from themed_control \
    import ThemedControl

#-------------------------------------------------------------------------
#  '_ThemedButtonEditor' class:
#-------------------------------------------------------------------------


class _ThemedButtonEditor(Editor):
    """ Traits UI themed button editor.
    """

    # The ThemedControl used for the button:
    button = Instance(ThemedControl)

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        # Create the button and its control:
        factory = self.factory
        label = factory.label
        if (label == '') and (factory.image is None):
            label = self.item.get_label(self.ui)
        label = self.string_value(label)

        self.button = button = ThemedControl(**factory.get(
            'theme', 'image', 'position', 'spacing')).set(
            text=label,
            controller=self,
            default_alignment='center',
            min_size=(80, 0))
        self.control = button.create_control(parent)

        # Set the tooltip:
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        pass

    #-- Trait Event Handlers -------------------------------------------------

    @on_trait_change('button.enabled')
    def _on_enabled_changed(self):
        """ Handles the button 'enabled' state changing.
        """
        if self.button.enabled:
            self.button.set(state='normal',
                            offset=(0, 0),
                            theme=self.factory.theme)
        else:
            self.button.set(state='disabled',
                            offset=(0, 0),
                            theme=self.factory.disabled_theme or
                            self.factory.theme)

    #-- ThemedControl Event Handlers -----------------------------------------

    def normal_left_down(self, x, y, event):
        if self.control.IsEnabled():
            self.button.set(state='down',
                            offset=(1, 1),
                            theme=self.factory.down_theme or
                            self.factory.theme)

    def normal_motion(self, x, y, event):
        hover = self.factory.hover_theme
        if self.control.IsEnabled() and (hover is not None):
            self.button.set(state='hover', theme=hover)
            self.control.CaptureMouse()

    def hover_left_down(self, x, y, event):
        self.control.ReleaseMouse()
        self.normal_left_down(x, y, event)

    def hover_motion(self, x, y, event):
        if not self.button.in_control(x, y):
            self.control.ReleaseMouse()
            self.button.set(state='normal', theme=self.factory.theme)

    def down_left_up(self, x, y, event):
        if self.button.in_control(x, y):
            self.value = True

            # If there is an associated view, display it:
            if self.factory.view is not None:
                self.object.edit_traits(view=self.factory.view,
                                        parent=self.control)

        self.button.set(state='normal',
                        offset=(0, 0),
                        theme=self.factory.theme)

    def down_motion(self, x, y, event):
        theme = self.factory.down_theme or self.factory.theme
        is_in = self.button.in_control(x, y)
        if not is_in:
            theme = self.factory.theme

        self.button.set(offset=(is_in, is_in), theme=theme)

#-------------------------------------------------------------------------
#  Create the editor factory object:
#-------------------------------------------------------------------------

# wxPython editor factory for themed button editors:


class ThemedButtonEditor(BasicEditorFactory):

    # The editor class to be created:
    klass = _ThemedButtonEditor

    # The button label:
    label = Str

    # The basic theme for the button (i.e. the 'up' state):
    theme = ATheme('@std:BG5')

    # The optional 'down' state theme for the button:
    down_theme = ATheme('@std:BE5')

    # The optional 'hover' state theme for the button:
    hover_theme = ATheme('@std:BG6')

    # The optional 'disabled' state theme for the button:
    disabled_theme = ATheme('@std:GG3')

    # The optional image to display in the button:
    image = Image

    # The position of the image relative to the text:
    position = Position

    # The amount of space between the image and the text:
    spacing = Spacing

    # The optional view to display when the button is clicked:
    view = AView



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   07/14/2007
#
#-------------------------------------------------------------------------

""" Defines the theme style information for a DockWindow and its components.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from pyface.ui_traits import Image
from traits.api import HasPrivateTraits, Bool

from .ui_traits import ATheme

#-------------------------------------------------------------------------
#  'DockWindowTheme' class:
#-------------------------------------------------------------------------


class DockWindowTheme(HasPrivateTraits):
    """ Defines the theme style information for a DockWindow and its components.
    """

    #-- Public Trait Definitions ---------------------------------------------

    # Use the theme background color as the DockWindow background color?
    use_theme_color = Bool(True)

    # Draw notebook tabs at the top (True) or the bottom (False)?
    tabs_at_top = Bool(True)

    # Active tab theme:
    tab_active = ATheme

    # Inactive tab theme:
    tab_inactive = ATheme

    # Optional image to use for right edge of rightmost inactive tab:
    tab_inactive_edge = Image

    # Tab hover theme (used for inactive tabs):
    tab_hover = ATheme

    # Optional image to use for right edge of rightmost hover tab:
    tab_hover_edge = Image

    # Tab background theme:
    tab_background = ATheme

    # Tab theme:
    tab = ATheme

    # Vertical splitter bar theme:
    vertical_splitter = ATheme

    # Horizontal splitter bar theme:
    horizontal_splitter = ATheme

    # Vertical drag bar theme:
    vertical_drag = ATheme

    # Horizontal drag bar theme:
    horizontal_drag = ATheme

#-------------------------------------------------------------------------
#  Define the default theme:
#-------------------------------------------------------------------------

# The current default DockWindow theme:
_dock_window_theme = None

# Gets/Sets the default DockWindow theme:


def dock_window_theme(theme=None):
    global _dock_window_theme

    if _dock_window_theme is None:
        from .default_dock_window_theme import default_dock_window_theme

        _dock_window_theme = default_dock_window_theme

    old_theme = _dock_window_theme
    if theme is not None:
        _dock_window_theme = theme

    return old_theme



from __future__ import absolute_import

from traits.testing.unittest_tools import unittest

from ..action import Action
from ..action_controller import ActionController
from ..action_event import ActionEvent


class TestActionController(unittest.TestCase):

    def setUp(self):
        # test whether function is called by updating list
        # XXX should really use mock
        self.memo = []

        def perform():
            self.memo.append('called')

        self.action = Action(name='Test', on_perform=perform)
        self.action_controller = ActionController()

    def test_perform(self):
        # test whether function is called by updating list
        # XXX should really use mock
        event = ActionEvent()
        self.action_controller.perform(self.action, event)
        self.assertEqual(self.memo, ['called'])

    def test_perform_none(self):
        action = Action(name='Test')
        event = ActionEvent()
        # does nothing, but shouldn't error
        self.action_controller.perform(action, event)

    def test_can_add_to_menu(self):
        result = self.action_controller.can_add_to_menu(self.action)
        self.assertTrue(result)

    def test_add_to_menu(self):
        # does nothing, shouldn't fail
        self.action_controller.add_to_menu(self.action)

    def test_can_add_to_toolbar(self):
        result = self.action_controller.can_add_to_toolbar(self.action)
        self.assertTrue(result)

    def test_add_to_toolbar(self):
        # does nothing, shouldn't fail
        self.action_controller.add_to_toolbar(self.action)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Enthought, Inc.
#
#------------------------------------------------------------------------------

""" Enthought pyface package component
"""

# Standard library imports.
import sys

# Major package imports.
import wx

# Enthought library imports.
from traits.api import HasTraits, Int, Property, provides, Tuple

# Local imports.
from pyface.i_system_metrics import ISystemMetrics, MSystemMetrics


@provides(ISystemMetrics)
class SystemMetrics(MSystemMetrics, HasTraits):
    """ The toolkit specific implementation of a SystemMetrics.  See the
    ISystemMetrics interface for the API documentation.
    """


    #### 'ISystemMetrics' interface ###########################################

    screen_width = Property(Int)

    screen_height = Property(Int)

    dialog_background_color = Property(Tuple)

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_screen_width(self):
        return wx.SystemSettings.GetMetric(wx.SYS_SCREEN_X)

    def _get_screen_height(self):
        return wx.SystemSettings.GetMetric(wx.SYS_SCREEN_Y)

    def _get_dialog_background_color(self):
        if sys.platform == 'darwin':
            # wx lies.
            color = wx.Colour(232, 232, 232)
        else:
            color = wx.SystemSettings_GetColour(wx.SYS_COLOUR_MENUBAR).Get()

        return (color[0]/255., color[1]/255., color[2]/255.)

#### EOF ######################################################################



""" Tests for the trait change event tracer. """
from traits.api import Float, HasTraits, on_trait_change
from traits.testing.unittest_tools import unittest

from traits import trait_notifiers


class FuzException(Exception):
    pass


class Foo(HasTraits):
    """ Test traits class with static and dynamic listeners.

    Changing `baz` triggers a dynamic listeners that modifies `bar`, which
    triggers one dynamic and one static listeners.
    """

    bar = Float
    baz = Float
    fuz = Float

    def _bar_changed(self):
        pass

    @on_trait_change('bar')
    def _on_bar_change_notification(self):
        pass

    @on_trait_change('baz')
    def _on_baz_change_notification(self):
        self.bar += 1

    @on_trait_change('fuz')
    def _on_fuz_change_notification(self):
        self.bar += 1
        raise FuzException('method')


class TestChangeEventTracers(unittest.TestCase):

    #### 'TestCase' protocol ##################################################

    def setUp(self):
        self.pre_change_events = []
        self.post_change_events = []
        self.exceptions = []
        trait_notifiers.push_exception_handler(
            lambda obj, name, old, new: None
        )

    def tearDown(self):
        trait_notifiers.pop_exception_handler()

    #### Private protocol #####################################################

    def _collect_pre_notification_events(self, *args):
        self.pre_change_events.append(args)

    def _collect_post_notification_events(self, *args, **kwargs):
        self.post_change_events.append(args)
        self.exceptions.extend(kwargs.values())

    #### Tests ################################################################

    def test_change_event_hooks(self):

        # Create the test object and a function listener.
        foo = Foo()

        def _on_foo_baz_changed(obj, name, old, new):
            pass
        foo.on_trait_change(_on_foo_baz_changed, 'baz')

        # Set the event tracer and trigger a cascade of change events.
        trait_notifiers.set_change_event_tracers(
            pre_tracer=self._collect_pre_notification_events,
            post_tracer=self._collect_post_notification_events)

        foo.baz = 3

        self.assertEqual(len(self.pre_change_events), 4)
        self.assertEqual(len(self.post_change_events), 4)

        expected_pre_events = [
            (foo, 'baz', 0.0, 3.0, foo._on_baz_change_notification),
            (foo, 'bar', 0.0, 1.0, foo._bar_changed.im_func),
            (foo, 'bar', 0.0, 1.0, foo._on_bar_change_notification),
            (foo, 'baz', 0.0, 3.0, _on_foo_baz_changed),
        ]
        self.assertEqual(self.pre_change_events, expected_pre_events)

        expected_post_events = [
            (foo, 'bar', 0.0, 1.0, foo._bar_changed.im_func),
            (foo, 'bar', 0.0, 1.0, foo._on_bar_change_notification),
            (foo, 'baz', 0.0, 3.0, foo._on_baz_change_notification),
            (foo, 'baz', 0.0, 3.0, _on_foo_baz_changed),
        ]
        self.assertEqual(self.post_change_events, expected_post_events)

        self.assertEqual(self.exceptions, [None] * 4)

        # Deactivate the tracer; it should not be called anymore.
        trait_notifiers.clear_change_event_tracers()
        foo.baz = 23
        self.assertEqual(len(self.pre_change_events), 4)
        self.assertEqual(len(self.post_change_events), 4)

    def test_change_event_hooks_after_exception(self):

        # Create the test object and a function listener.
        foo = Foo()

        def _on_foo_fuz_changed(obj, name, old, new):
            raise FuzException('function')
        foo.on_trait_change(_on_foo_fuz_changed, 'fuz')

        # Set the event tracer and trigger a cascade of change events.
        trait_notifiers.set_change_event_tracers(
            pre_tracer=self._collect_pre_notification_events,
            post_tracer=self._collect_post_notification_events)

        foo.fuz = 3

        self.assertEqual(len(self.pre_change_events), 4)
        self.assertEqual(len(self.post_change_events), 4)

        expected_pre_events = [
            (foo, 'fuz', 0.0, 3.0, foo._on_fuz_change_notification),
            (foo, 'bar', 0.0, 1.0, foo._bar_changed.im_func),
            (foo, 'bar', 0.0, 1.0, foo._on_bar_change_notification),
            (foo, 'fuz', 0.0, 3.0, _on_foo_fuz_changed),
        ]
        self.assertEqual(self.pre_change_events, expected_pre_events)

        expected_post_events = [
            (foo, 'bar', 0.0, 1.0, foo._bar_changed.im_func),
            (foo, 'bar', 0.0, 1.0, foo._on_bar_change_notification),
            (foo, 'fuz', 0.0, 3.0, foo._on_fuz_change_notification),
            (foo, 'fuz', 0.0, 3.0, _on_foo_fuz_changed),
        ]
        self.assertEqual(self.post_change_events, expected_post_events)

        self.assertEqual(self.exceptions[:2], [None, None])
        self.assertIsInstance(self.exceptions[2], FuzException)
        self.assertEqual(self.exceptions[2].args, ('method',))
        self.assertIsInstance(self.exceptions[3], FuzException)
        self.assertEqual(self.exceptions[3].args, ('function',))


if __name__ == '__main__':
    unittest.main()



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought logger package component>
#------------------------------------------------------------------------------

import logging

from apptools.preferences.ui.api import PreferencesPage
from traits.api import Bool, Trait, Str
from traitsui.api import EnumEditor, Group, Item, View

class LoggerPreferencesPage(PreferencesPage):
    """ A preference page for the logger plugin.
    """

    #### 'PreferencesPage' interface ##########################################

    # The page's category (e.g. 'General/Appearance'). The empty string means
    # that this is a top-level page.
    category = ''

    # The page's help identifier (optional). If a help Id *is* provided then
    # there will be a 'Help' button shown on the preference page.
    help_id = ''

    # The page name (this is what is shown in the preferences dialog.
    name = 'Logger'

    # The path to the preferences node that contains the preferences.
    preferences_path = 'apptools.logger'


    #### Preferences ###########################################################

    # The log levels
    level = Trait('Info',
        {'Debug'    : logging.DEBUG,
         'Info'     : logging.INFO,
         'Warning'  : logging.WARNING,
         'Error'    : logging.ERROR,
         'Critical' : logging.CRITICAL,
        },
        is_str = True,
    )

    enable_agent = Bool(False)
    smtp_server = Str
    to_address = Str
    from_address = Str


    # The view used to change the plugin preferences
    traits_view = View(
        Group(
            Group(
                Item(
                    name='level',
                    editor=EnumEditor(
                        values={
                            'Debug'    : '1:Debug',
                            'Info'     : '2:Info',
                            'Warning'  : '3:Warning',
                            'Error'    : '4:Error' ,
                            'Critical' : '5:Critical',
                        },
                    ),
                    style='simple',
                ),
                label='Logger Settings',
                show_border=True,
            ),
            Group(Item(name='10')),
            Group(
                Group(
                    Group(Item(name='enable_agent', label='Enable quality agent'), show_left=False),
                    Group(Item(name='smtp_server', label='SMTP server'),
                          Item(name='from_address'),
                          Item(name='to_address'), enabled_when='enable_agent==True')),
                label='Quality Agent Settings',
                show_border=True,
            ),
        ),
    )


#### EOF ######################################################################



from __future__ import absolute_import

from traits.testing.unittest_tools import unittest

from ..gui import GUI
from ..image_resource import ImageResource
from ..splash_screen import SplashScreen
from ..toolkit import toolkit_object

GuiTestAssistant = toolkit_object('util.gui_test_assistant:GuiTestAssistant')
no_gui_test_assistant = (GuiTestAssistant.__name__ == 'Unimplemented')


@unittest.skipIf(no_gui_test_assistant, 'No GuiTestAssistant')
class TestWindow(unittest.TestCase, GuiTestAssistant):

    def setUp(self):
        GuiTestAssistant.setUp(self)
        self.window = SplashScreen()

    def tearDown(self):
        if self.window.control is not None:
            with self.delete_widget(self.window.control):
                self.widnow.destroy()
        GuiTestAssistant.tearDown(self)

    def test_destroy(self):
        # test that destroy works even when no control
        self.window.destroy()

    def test_open_close(self):
        # test that opening and closing works as expected
        with self.assertTraitChanges(self.window, 'opening', count=1):
            with self.assertTraitChanges(self.window, 'opened', count=1):
                self.window.open()
        self.gui.process_events()
        with self.assertTraitChanges(self.window, 'closing', count=1):
            with self.assertTraitChanges(self.window, 'closed', count=1):
                self.window.close()
        self.gui.process_events()

    def test_show(self):
        # test that show works as expected
        self.window._create()
        self.window.show(True)
        self.gui.process_events()
        self.window.show(False)
        self.gui.process_events()
        self.window.destroy()

    def test_image(self):
        # test that images work
        self.window.image = ImageResource('core')
        with self.assertTraitChanges(self.window, 'opening', count=1):
            with self.assertTraitChanges(self.window, 'opened', count=1):
                self.window.open()
        self.gui.process_events()
        with self.assertTraitChanges(self.window, 'closing', count=1):
            with self.assertTraitChanges(self.window, 'closed', count=1):
                self.window.close()
        self.gui.process_events()

    def test_text(self):
        # test that images work
        self.window.text = "Splash screen"
        with self.assertTraitChanges(self.window, 'opening', count=1):
            with self.assertTraitChanges(self.window, 'opened', count=1):
                self.window.open()
        self.gui.process_events()
        with self.assertTraitChanges(self.window, 'closing', count=1):
            with self.assertTraitChanges(self.window, 'closed', count=1):
                self.window.close()
        self.gui.process_events()

    def test_text_changed(self):
        # test that images work
        # XXX this throws a non-failing exception on wx
        #     - probably the way the test is written.
        with self.assertTraitChanges(self.window, 'opening', count=1):
            with self.assertTraitChanges(self.window, 'opened', count=1):
                self.window.open()
        self.gui.process_events()
        self.window.text = "Splash screen"
        self.gui.process_events()
        with self.assertTraitChanges(self.window, 'closing', count=1):
            with self.assertTraitChanges(self.window, 'closed', count=1):
                self.window.close()
        self.gui.process_events()



'''
Sqlite3 based file system using PyFileSystem

Developed and Contributed by Sensible Softwares Pvt. Ltd.
(http://bootstraptoday.com)

Contributed under the terms of the BSD License:
http://www.opensource.org/licenses/bsd-license.php
'''

import tempfile
import datetime

from fs.path import iteratepath, normpath,dirname,forcedir
from fs.path import frombase, basename,pathjoin
from fs.base import *
from fs.errors import *
from fs import _thread_synchronize_default
import apsw

def fetchone(cursor):
    '''
    return a single row from the cursor (equivalent to pysqlite fetchone function)
    '''
    row = None
    try:
        row = cursor.next()
    except:
        pass
    return(row)

def remove_end_slash(dirname):
    if dirname.endswith('/'):
        return dirname[:-1]
    return dirname

class SqliteFsFileBase(object):
    '''
    base class for representing the files in the sqlite file system
    '''
    def __init__(self, fs, path, id, real_file=None):
        assert(fs != None)
        assert(path != None)
        assert(id != None)
        self.fs = fs
        self.path = path
        self.id = id
        self.closed = False
        #real file like object. Most of the methods are passed to this object
        self.real_stream= real_file
        
    def close(self):
        if not self.closed and self.real_stream is not None:
            self._do_close()
            self.fs._on_close(self)
            self.real_stream.close()
            self.closed = True
        
    def __str__(self):
        return "<SqliteFS File in %s %s>" % (self.fs, self.path)

    __repr__ = __str__

    def __unicode__(self):
        return u"<SqliteFS File in %s %s>" % (self.fs, self.path)

    def __del__(self):
        if not self.closed:
            self.close()

    def flush(self):
        self.real_stream.flush()
        
    def __iter__(self):
        raise OperationFailedError('__iter__', self.path)
        
    def next(self):
        raise OperationFailedError('next', self.path)
        
    def readline(self, *args, **kwargs):
        raise OperationFailedError('readline', self.path)        
    
    def read(self, size=None):
        raise OperationFailedError('read', self.path)                

    def seek(self, *args, **kwargs):
        return self.real_stream.seek(*args, **kwargs)

    def tell(self):
        return self.real_stream.tell()

    def truncate(self, *args, **kwargs):
        raise OperationFailedError('truncate', self.path)
        
    def write(self, data):
        raise OperationFailedError('write', self.path)
        
    def writelines(self, *args, **kwargs):
        raise OperationFailedError('writelines', self.path)
        
    def __enter__(self):
        return self

    def __exit__(self,exc_type,exc_value,traceback):
        self.close()
        return False
    
class SqliteWritableFile(SqliteFsFileBase):
    '''
    represents an sqlite file. Usually used for 'writing'. OnClose will
    actually 'copy the contents from temp disk file to sqlite blob
    '''
    def __init__(self,fs, path, id):
        super(SqliteWritableFile, self).__init__(fs, path, id)
        #open a temp file and return that.
        self.real_stream = tempfile.SpooledTemporaryFile(max_size='128*1000')
        
    def _do_close(self):
        #push the contents of the file to blob
        self.fs._writeblob(self.id, self.real_stream)
        
    def truncate(self, *args, **kwargs):
        return self.real_stream.truncate(*args, **kwargs)
        
    def write(self, data):
        return self.real_stream.write(data)

    def writelines(self, *args, **kwargs):
        return self.real_stream.writelines(*args, **kwargs)
        
class SqliteReadableFile(SqliteFsFileBase):
    def __init__(self,fs, path, id, real_file):
        super(SqliteReadableFile, self).__init__(fs, path, id, real_file)
        assert(self.real_stream != None)        
        
    def _do_close(self):
        pass
    
    def __iter__(self):
        return iter(self.real_stream)

    def next(self):
        return self.real_stream.next()

    def readline(self, *args, **kwargs):
        return self.real_stream.readline(*args, **kwargs)
    
    def read(self, size=None):
        if( size==None):
            size=-1
        return self.real_stream.read(size)

            
class SqliteFS(FS):
    '''
    sqlite based file system to store the files in sqlite database as 'blobs'
    We need two tables one to store file or directory meta data
    another for storing the file contain. Two are seperate so that same file
    can be refered from multiple directories.
    FsFileMetaData table :
        id : file id
        name : name of file
        parent : id of parent directory for the file.
        
    FsDirMetaData table:
        name : name of the directory (wihtout parent directory names)
        fullpath : full path of the directory including the parent directory name
        parent_id : id of the parent directory
                
    FsFileTable:
        size : file size in bytes (this is actual file size). Blob size may be
           different if compressed
        type : file type (extension or mime type)
        compression : For future use, Initially None. Later it will define type
            of compression used to compress the file
        last_modified : timestamp of last modification
        author : who changed it last
        content : blob where actual file contents are stored.
        
    TODO : Need an open files table or a flag in sqlite database. To avoid
    opening the file twice. (even from the different process or thread)
    '''
    
    def __init__(self, sqlite_filename):
        super(SqliteFS, self).__init__()
        self.dbpath =sqlite_filename
        self.dbcon =None        
        self.__actual_query_cur = None
        self.__actual_update_cur =None
        self.open_files = []
        
    def close(self):
        '''
        unlock all files. and close all open connections.
        '''
        self.close_all()        
        self._closedb()
        super(SqliteFS,self).close()
        
    def _initdb(self):
        if( self.dbcon is None):
            self.dbcon = apsw.Connection(self.dbpath)        
            self._create_tables()
    
    @property
    def _querycur(self):
        assert(self.dbcon != None)
        if( self.__actual_query_cur == None):
            self.__actual_query_cur = self.dbcon.cursor()
        return(self.__actual_query_cur)
    
    @property
    def _updatecur(self):
        assert(self.dbcon != None)
        if( self.__actual_update_cur == None):
            self.__actual_update_cur = self.dbcon.cursor()
        return(self.__actual_update_cur)
        
    def _closedb(self):
        self.dbcon.close()
        
    def close_all(self):
        '''
        close all open files
        '''
        openfiles = list(self.open_files)
        for fileobj in openfiles:
            fileobj.close()
        
    def _create_tables(self):
        cur = self._updatecur
        cur.execute("CREATE TABLE IF NOT EXISTS FsFileMetaData(name text, fileid INTEGER, parent INTEGER)")
        cur.execute("CREATE TABLE IF NOT EXISTS FsDirMetaData(name text, fullpath TEXT, parentid INTEGER,\
                    created timestamp)")
        cur.execute("CREATE TABLE IF NOT EXISTS FsFileTable(type text, compression text, author TEXT, \
                    created timestamp, last_modified timestamp, last_accessed timestamp, \
                    locked BOOL, size INTEGER, contents BLOB)")
        
        #if the root directory name is created
        rootid = self._get_dir_id('/')
        if( rootid is None):
            cur.execute("INSERT INTO FsDirMetaData (name, fullpath) VALUES ('/','/')")
            
    def _get_dir_id(self, dirpath):
        '''
        get the id for given directory path.
        '''
        dirpath = remove_end_slash(dirpath)
        if( dirpath== None or len(dirpath)==0):
            dirpath = '/'
            
        self._querycur.execute("SELECT rowid from FsDirMetaData where fullpath=?",(dirpath,))
        dirid = None
        dirrow = fetchone(self._querycur)
        if( dirrow):
            dirid = dirrow[0]
        
        return(dirid)
        
    def _get_file_id(self, dir_id, filename):
        '''
        get the file id from the path
        '''
        assert(dir_id != None)
        assert(filename != None)
        file_id = None
        self._querycur.execute("select rowid from FsFileMetaData where name=? and parent=?",(filename,dir_id))
        row = fetchone(self._querycur)
        if( row ):
            file_id = row[0]
        return(file_id)
        
    def _get_file_contentid(self, file_id):
        '''
        return the file content id from the 'content' table (i.e. FsFileTable)        
        '''
        assert(file_id != None)
        content_id = None
        self._querycur.execute("select fileid from FsFileMetaData where ROWID=?",(file_id,))
        row = fetchone(self._querycur)
        assert(row != None)
        content_id = row[0]
        return(content_id)
        
    def _create_file_entry(self, dirid, filename, **kwargs):
        '''
        create file entry in the file table
        '''
        assert(dirid != None)
        assert(filename != None)
        #insert entry in file metadata table
        author = kwargs.pop('author', None)
        created = datetime.datetime.now().isoformat()
        last_modified = created
        compression = 'raw'
        size = 0
        self._updatecur.execute("INSERT INTO FsFileTable(author, compression, size, created, last_modified) \
                    values(?, ?, ?, ?, ?)",(author, compression, size, created, last_modified))
        content_id = self.dbcon.last_insert_rowid()
        #insert entry in file table
        self._updatecur.execute("INSERT INTO FsFileMetaData(name, parent, fileid) VALUES(?,?,?)",(filename, dirid, content_id))
        #self.dbcon.commit()
        fileid = self.dbcon.last_insert_rowid()
        return(fileid)
            
    def _writeblob(self, fileid, stream):
        '''
        extract the data from stream and write it as blob.
        '''
        size = stream.tell()
        last_modified = datetime.datetime.now().isoformat()
        self._updatecur.execute('UPDATE FsFileTable SET size=?, last_modified=?, contents=? where rowid=?',
                (size, last_modified, apsw.zeroblob(size), fileid))
        blob_stream=self.dbcon.blobopen("main", "FsFileTable", "contents", fileid, True) # 1 is for read/write
        stream.seek(0)
        blob_stream.write(stream.read())
        blob_stream.close()        
        
    def _on_close(self, fileobj):        
        #Unlock file on close.
        assert(fileobj != None and fileobj.id != None)
        self._lockfileentry(fileobj.id, lock=False)
        #Now remove it from openfile list.
        self.open_files.remove(fileobj)
            
    def _islocked(self, fileid):
        '''
        check if the file is locked.
        '''
        locked=False
        if( fileid):
            content_id = self._get_file_contentid(fileid)
            assert(content_id != None)
            self._querycur.execute("select locked from FsFileTable where rowid=?",(content_id,))
            row = fetchone(self._querycur)
            assert(row != None)
            locked = row[0]
        return(locked)
        
    def _lockfileentry(self, contentid, lock=True):
        '''
        lock the file entry in the database.
        '''
        assert(contentid != None)
        last_accessed=datetime.datetime.now().isoformat()
        self._updatecur.execute('UPDATE FsFileTable SET locked=?, last_accessed=? where rowid=?',
                    (lock, last_accessed, contentid))
        
    def _makedir(self, parent_id, dname):        
        self._querycur.execute("SELECT fullpath from FsDirMetaData where rowid=?",(parent_id,))
        row = fetchone(self._querycur)
        assert(row != None)
        parentpath = row[0]        
        fullpath= pathjoin(parentpath, dname)
        fullpath= remove_end_slash(fullpath)        
        created = datetime.datetime.now().isoformat()
        self._updatecur.execute('INSERT INTO FsDirMetaData(name, fullpath, parentid,created) \
                    VALUES(?,?,?,?)', (dname, fullpath, parent_id,created))
        
    def _rename_file(self, src, dst):
        '''
        rename source file 'src' to destination file 'dst'
        '''
        srcdir = dirname(src)
        srcfname = basename(src)
        dstdir = dirname(dst)
        dstfname = basename(dst)
        #Make sure that the destination directory exists and destination file
        #doesnot exist.
        dstdirid = self._get_dir_id(dstdir)
        if( dstdirid == None):
            raise ParentDirectoryMissingError(dst)
        dstfile_id = self._get_file_id(dstdirid, dstfname)
        if( dstfile_id != None):
            raise DestinationExistsError(dst)
        #All checks are done. Delete the entry for the source file.
        #Create an entry for the destination file.            
        
        srcdir_id = self._get_dir_id(srcdir)
        assert(srcdir_id != None)
        srcfile_id = self._get_file_id(srcdir_id, srcfname)
        assert(srcfile_id != None)
        srccontent_id = self._get_file_contentid(srcfile_id)
        self._updatecur.execute('DELETE FROM FsFileMetaData where ROWID=?',(srcfile_id,))
        self._updatecur.execute("INSERT INTO FsFileMetaData(name, parent, fileid) \
                            VALUES(?,?,?)",(dstfname, dstdirid, srccontent_id))
        
    def _rename_dir(self, src, dst):
        src = remove_end_slash(src)
        dst = remove_end_slash(dst)
        dstdirid = self._get_dir_id(dst)
        if( dstdirid != None):
            raise DestinationExistsError(dst)
        dstparent = dirname(dst)
        dstparentid = self._get_dir_id(dstparent)
        if(dstparentid == None):
            raise ParentDirectoryMissingError(dst)
        srcdirid = self._get_dir_id(src)
        assert(srcdirid != None)
        dstdname = basename(dst)        
        self._updatecur.execute('UPDATE FsDirMetaData SET name=?, fullpath=?, \
                    parentid=? where ROWID=?',(dstdname, dst, dstparentid, srcdirid,))
        
    def _get_dir_list(self, dirid, path, full):
        assert(dirid != None)
        assert(path != None)
        if( full==True):
            dirsearchpath = path + r'%'
            self._querycur.execute('SELECT fullpath FROM FsDirMetaData where fullpath LIKE ?',
                                   (dirsearchpath,))            
        else:
            #search inside this directory only
            self._querycur.execute('SELECT fullpath FROM FsDirMetaData where parentid=?',
                                   (dirid,))
        dirlist = [row[0] for row in self._querycur]        
        return dirlist
            
    def _get_file_list(self, dirpath, full):
        assert(dirpath != None)
                
        if( full==True):
            searchpath = dirpath + r"%"
            self._querycur.execute('SELECT FsFileMetaData.name, FsDirMetaData.fullpath \
                FROM FsFileMetaData, FsDirMetaData where FsFileMetaData.parent=FsDirMetaData.ROWID \
                    and FsFileMetaData.parent in (SELECT rowid FROM FsDirMetaData \
                    where fullpath LIKE ?)',(searchpath,))
        else:
            parentid = self._get_dir_id(dirpath)
            self._querycur.execute('SELECT FsFileMetaData.name, FsDirMetaData.fullpath \
                FROM FsFileMetaData, FsDirMetaData where FsFileMetaData.parent=FsDirMetaData.ROWID \
                    and FsFileMetaData.parent =?',(parentid,))
            
        filelist = [pathjoin(row[1],row[0]) for row in self._querycur]        
        return(filelist)
        
    def _get_dir_info(self, path):
        '''
        get the directory information dictionary.
        '''
        info = dict()
        info['st_mode'] = 0755
        return info
        
    def _get_file_info(self, path):
        filedir = dirname(path)
        filename = basename(path)
        dirid = self._get_dir_id(filedir)
        assert(dirid is not None)
        fileid = self._get_file_id(dirid, filename)
        assert(fileid is not None)
        contentid = self._get_file_contentid(fileid)
        assert(contentid is not None)
        self._querycur.execute('SELECT author, size, created, last_modified, last_accessed \
                        FROM FsFileTable where rowid=?',(contentid,))
        row = fetchone(self._querycur)
        assert(row != None)
        info = dict()
        info['author'] = row[0]
        info['size'] = row[1]
        info['created'] = row[2]
        info['last_modified'] = row[3]
        info['last_accessed'] = row[4]
        info['st_mode'] = 0666
        return(info)
        
    def _isfile(self,path):
        path = normpath(path)
        filedir = dirname(path)
        filename = basename(path)
        dirid = self._get_dir_id(filedir)        
        return(dirid is not None and self._get_file_id(dirid, filename) is not None)
        
    def _isdir(self,path):
        path = normpath(path)        
        return(self._get_dir_id(path) is not None)
        
    def _isexist(self,path):
        return self._isfile(path) or self._isdir(path)
        
    @synchronize
    def open(self, path, mode='r', **kwargs):
        self._initdb()
        path = normpath(path)
        filedir = dirname(path)
        filename = basename(path)
        
        dir_id = self._get_dir_id(filedir)
        if( dir_id == None):
            raise ResourceNotFoundError(filedir)
            
        file_id = self._get_file_id(dir_id, filename)        
        if( self._islocked(file_id)):
                raise ResourceLockedError(path)            
            
        sqfsfile=None
        if 'r' in mode:
            if file_id is None:
                raise ResourceNotFoundError(path)
            content_id = self._get_file_contentid(file_id)
            #make sure lock status is updated before the blob is opened
            self._lockfileentry(content_id, lock=True)
            blob_stream=self.dbcon.blobopen("main", "FsFileTable", "contents", file_id, False) # 1 is for read/write
            sqfsfile = SqliteReadableFile(self, path, content_id, blob_stream)            
                    
        elif 'w' in mode or 'a' in mode:
            if( file_id is None):
                file_id= self._create_file_entry(dir_id, filename)
                assert(file_id != None)
            
            content_id = self._get_file_contentid(file_id)
            #file_dir_entry.accessed_time = datetime.datetime.now()                        
            self._lockfileentry(content_id, lock=True)
            sqfsfile = SqliteWritableFile(self, path, content_id)            
            
        if( sqfsfile):
            self.open_files.append(sqfsfile)
            return sqfsfile
        
        raise ResourceNotFoundError(path)        
    
    @synchronize
    def isfile(self, path):
        self._initdb()
        return self._isfile(path)
        
    @synchronize
    def isdir(self, path):
        self._initdb()
        return self._isdir(path)
    
    @synchronize
    def listdir(self, path='/', wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
        path = normpath(path)        
        dirid = self._get_dir_id(path)
        if( dirid == None):
            raise ResourceInvalidError(path)
        
        dirlist = self._get_dir_list(dirid, path,full)
        if( dirs_only):
            pathlist = dirlist
        else:            
            filelist = self._get_file_list(path, full)
            
            if( files_only == True):
                pathlist = filelist
            else:
                pathlist = filelist + dirlist
                
            
        if( wildcard and dirs_only == False):
            pass
        
        if( absolute == False):
            pathlist = map(lambda dpath:frombase(path,dpath), pathlist)
            
        return(pathlist)
        
    
    @synchronize
    def makedir(self, path, recursive=False, allow_recreate=False):
        self._initdb()
        path = remove_end_slash(normpath(path))
            
        if(self._isexist(path)==False):
            parentdir = dirname(path)
            dname = basename(path)
            
            parent_id = self._get_dir_id(parentdir)
            if( parent_id ==None):
                if( recursive == False):                
                    raise ParentDirectoryMissingError(path)
                else:
                    self.makedir(parentdir, recursive,allow_recreate)
                    parent_id = self._get_dir_id(parentdir)
            self._makedir(parent_id,dname)
        else:
            raise DestinationExistsError(path)
    
    @synchronize
    def remove(self, path):
        self._initdb()
        path = normpath(path)
        if( self.isdir(path)==True):
            #path is actually a directory
            raise ResourceInvalidError(path)
        
        filedir = dirname(path)
        filename = basename(path)
        dirid = self._get_dir_id(filedir)
        fileid = self._get_file_id(dirid, filename)
        if( fileid == None):
            raise ResourceNotFoundError(path)
        
        content_id = self._get_file_contentid(fileid)
        
        self._updatecur.execute("DELETE FROM FsFileMetaData where ROWID=?",(fileid,))
        #check there is any other file pointing to same location. If not
        #delete the content as well.
        self._querycur.execute('SELECT count(*) FROM FsFileMetaData where fileid=?',
                    (content_id,))
        row = fetchone(self._querycur)
        if( row == None or row[0] == 0):
            self._updatecur.execute("DELETE FROM FsFileTable where ROWID=?",(content_id,))            
    
    @synchronize
    def removedir(self,path, recursive=False, force=False):
        self._initdb()
        path = normpath(path)
        if( self.isfile(path)==True):
            #path is actually a file
            raise ResourceInvalidError(path)
        dirid = self._get_dir_id(path)
        if( dirid == None):
            raise ResourceNotFoundError(path)
        #check if there are any files in this directory
        self._querycur.execute("SELECT COUNT(*) FROM FsFileMetaData where parent=?",(dirid,))
        row = fetchone(self._qurycur)
        if( row[0] > 0):
            raise DirectoryNotEmptyError(path)
        self._updatecur.execute("DELETE FROM FsDirMetaData where ROWID=?",(dirid,))            
    
    @synchronize
    def rename(self,src, dst):
        self._initdb()
        src = normpath(src)
        dst = normpath(dst)
        if self._isexist(dst)== False:
            #first check if this is a directory rename or a file rename        
            if( self.isfile(src)):
                self._rename_file(src, dst)
            elif self.isdir(src):
                self._rename_dir(src, dst)
            else:
                raise ResourceNotFoundError(path)
        else:            
            raise DestinationExistsError(dst)
            
    @synchronize
    def getinfo(self, path):
        self._initdb()        
        path = normpath(path)
        isfile = False
        isdir = self.isdir(path)
        if( isdir == False):
            isfile=self.isfile(path)
        
        if( not isfile and not isdir):
            raise ResourceNotFoundError(path)
                    
        if isdir:
            info= self._get_dir_info(path)
        else:
            info= self._get_file_info(path)
        return(info)

#import msvcrt # built-in module
#
#def kbfunc():
#    return ord(msvcrt.getch()) if msvcrt.kbhit() else 0
#
#def mount_windows(sqlfilename, driveletter):
#    sqfs = SqliteFS(sqlfilename)
#    from fs.expose import dokan
#    #mp = dokan.mount(sqfs,driveletter,foreground=True)
#    #mp.unmount()
#    sqfs.close()
#    
#def run_tests(sqlfilename):
#    fs = SqliteFS(sqlfilename)
#    fs.makedir('/test')
#    f = fs.open('/test/test.txt', "w")
#    f.write("This is a test")
#    f.close()
#    f = fs.open('/test/test.txt', "r")
#    contents = f.read()
#    print contents
#    f.close()
#    print "testing file rename"
#    fs.rename('/test/test.txt', '/test/test1.txt')
#    f = fs.open('/test/test1.txt', "r")
#    print contents
#    f.close()
#    print "done testing file rename"
#    print "testing directory rename"
#    fs.rename('/test', '/test1')
#    f = fs.open('/test1/test1.txt', "r")
#    contents = f.read()
#    print contents
#    f.close()
#    print "done testing directory rename"
#    flist = fs.listdir('/', full=True,absolute=True,files_only=True)
#    print flist
#    fs.close()
#    
#if __name__ == '__main__':
#    run_tests("sqfs.sqlite")
#    mount_windows("sqfs.sqlite", 'm')
#    
#    #fs.remove('/test1/test1.txt')
#    #try:
#    #    f = fs.open('/test1/test1.txt', "r")
#    #except ResourceNotFoundError:
#    #    print "Success : file doesnot exist"
#    #fs.browse()
#    


import OpenGL.GL as gl


class CurrentTransform:
    def __enter__(self):
        gl.glPushMatrix()
        return self

    def __exit__(self, type, value, traceback):
        gl.glPopMatrix()


class DisplayList:
    def __init__(self):
        self.display_list_id = gl.glGenLists(1)

    def __call__(self):
        gl.glCallList(self.display_list_id)

class DisplayListGenerator:
    def __enter__(self):
        dl = DisplayList()
        gl.glNewList(dl.display_list_id, gl.GL_COMPILE)
        return dl

    def __exit__(self, type, value, traceback):
        gl.glEndList()

class ViewportAs:
    def __init__(self, x, y, w, h, viewport_w=None, viewport_h=None, invert_y=False):
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.viewport_w = viewport_w if viewport_w else w
        self.viewport_h = viewport_h if viewport_h else h
        self.invert_y = False

    def __enter__(self):
        self.old_viewport = list(gl.glGetFloatv(gl.GL_VIEWPORT))

        gl.glViewport(0, 0, self.viewport_w, self.viewport_h)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPushMatrix()
        gl.glLoadIdentity()
        if not self.invert_y:
            gl.glOrtho(self.x, self.w, self.y, self.h, 0, 1)
        else:
            gl.glOrtho(self.x, self.w, self.h, self.y, 0, 1)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()
        return self

    def __exit__(self, type, value, traceback):
        viewport = self.old_viewport
        gl.glMatrixMode(gl.GL_PROJECTION)

        gl.glPopMatrix()
        gl.glViewport(int(viewport[0]), int(viewport[1]), int(viewport[2]), int(viewport[3]))
        gl.glMatrixMode(gl.GL_MODELVIEW)


#!/usr/bin/env python

"""
Utilities to help with grid io

NOTE: this isn't used yet, but should be useful for loading non
UGRID-compliant files.

"""

from __future__ import (absolute_import, division, print_function)

import netCDF4
import numpy as np

from ..ugrid import UGrid


def load_from_varnames(filename, names_mapping, attribute_check=None):
    """
    Load a UGrid from a netcdf file where the roles are defined by the
    names of the variables.

    :param filename: names of the file to load (or OPeNDAP URL).

    :param names_mapping: dict that maps the variable names to UGrid components

    :param attribute_check=None: list of global attributes that are expected
    :type attribute_check: list of tuples to check. Example:
                           [('grid_type','triangular'),] will check if the
                           grid_type attribute is set to "triangular"

    The names_mapping dict has to contain at least: 'nodes_lon', 'nodes_lat'

    Optionally (and mostly required), it can contain: face_face_connectivity',
    'face_coordinates_lon', 'face_coordinates_lat', and 'faces'

    """
    ug = UGrid()
    attribute_check = {} if attribute_check is None else attribute_check

    nc = netCDF4.Dataset(filename)

    # Check for the specified attributes.
    for name, value in attribute_check:
        if nc.getncattr(name).lower() != value:
            raise ValueError('This does not appear to be a valid file:\n'
                             'It does not have the "{}"="{}"'
                             'global attribute set'.format(name, value))

    # Nodes.
    lon = nc.variables[names_mapping['nodes_lon']]
    lat = nc.variables[names_mapping['nodes_lat']]

    num_nodes = lon.size
    ug.nodes = np.zeros((num_nodes, 2), dtype=lon.dtype)
    ug.nodes[:, 0] = lon[:]
    ug.nodes[:, 1] = lat[:]

    # Faces.
    faces = nc.variables[names_mapping['faces']]
    # FIXME: This logic assumes there are more than three triangles.
    if faces.shape[0] <= faces.shape[1]:
        # Fortran order.
        faces = faces[:].T
    else:
        faces = faces[:]

    # One-indexed?
    if faces.min() == 1:
        one_indexed = True
    else:
        one_indexed = False

    if one_indexed:
        faces -= 1
    ug.faces = faces

    # Connectivity (optional).
    if 'face_face_connectivity' in names_mapping:
        face_face_connectivity = nc.variables[names_mapping['face_face_connectivity']]  # noqa
        # FIXME: This logic assumes there are more than three triangles.
        if face_face_connectivity.shape[0] <= face_face_connectivity.shape[1]:
            # Fortran order.
            face_face_connectivity = face_face_connectivity[:].T
        else:
            face_face_connectivity = face_face_connectivity[:]
        if one_indexed:
            face_face_connectivity -= 1
        ug.face_face_connectivity = face_face_connectivity

    # Center (optional).
    if ('face_coordinates_lon' in names_mapping and
       'face_coordinates_lat' in names_mapping):

        ug.face_coordinates = np.zeros((len(ug.faces), 2), dtype=lon.dtype)
        ug.face_coordinates[:, 0] = nc.variables[names_mapping['face_coordinates_lon']][:]  # noqa
        ug.face_coordinates[:, 1] = nc.variables[names_mapping['face_coordinates_lat']][:]  # noqa

    # Boundaries (optional).
    if 'boundaries' in names_mapping:
        # FIXME: this one is weird and non-conforming!
        # Ignoring the second two fields. What are they?
        boundaries = nc.variables[names_mapping['boundaries']][:, :2]
        if one_indexed:
            boundaries -= 1
        ug.boundaries = boundaries

    return ug



# Import the toolkit specific version.
from pyface.toolkit import toolkit_object
SplitEditorAreaPane = toolkit_object('tasks.split_editor_area_pane:'
                                     'SplitEditorAreaPane')

EditorAreaWidget = toolkit_object('tasks.split_editor_area_pane:'
                                    'EditorAreaWidget')



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought application scripting package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import Any, HasTraits, provides, Str

# Local imports.
from i_bind_event import IBindEvent


@provides(IBindEvent)
class BindEvent(HasTraits):
    """The default implementation of the bind event interface."""



    #### 'IBindEvent' interface ###############################################

    # This is the name being bound or unbound.
    name = Str

    # This is the object being bound to the name.  It is None if the name is
    # being unbound.
    obj = Any



"""
fs.filelike
===========

This module takes care of the groundwork for implementing and manipulating
objects that provide a rich file-like interface, including reading, writing,
seeking and iteration.

The main class is FileLikeBase, which implements the entire file-like interface
on top of primitive _read(), _write(), _seek(), _tell() and _truncate() methods.
Subclasses may implement any or all of these methods to obtain the related
higher-level file behaviors.

Other useful classes include:

    * StringIO:   a version of the builtin StringIO class, patched to more
                  closely preserve the semantics of a standard file.

    * FileWrapper:  a generic base class for wrappers around a filelike object
                    (think e.g. compression or decryption).

    * SpooledTemporaryFile:  a version of the builtin SpooledTemporaryFile
                             class, patched to more closely preserve the
                             semantics of a standard file.

    * LimitBytesFile:  a filelike wrapper that limits the total bytes read
                       from a file; useful for turning a socket into a file
                       without reading past end-of-data.

"""
# Copyright (C) 2006-2009, Ryan Kelly
# All rights reserved; available under the terms of the MIT License.

import tempfile as _tempfile

import fs


class NotReadableError(IOError):
    pass
class NotWritableError(IOError):
    pass
class NotSeekableError(IOError):
    pass
class NotTruncatableError(IOError):
    pass

import six
from six import PY3, b

if PY3:
    from six import BytesIO as _StringIO
else:
    try:
        from cStringIO import StringIO as _StringIO
    except ImportError:
        from StringIO import StringIO as _StringIO


class FileLikeBase(object):
    """Base class for implementing file-like objects.

    This class takes a lot of the legwork out of writing file-like objects
    with a rich interface.  It implements the higher-level file-like methods
    on top of five primitive methods: _read, _write, _seek, _tell and
    _truncate. See their docstrings for precise details on how these methods
    behave.

    Subclasses then need only implement some subset of these methods for
    rich file-like interface compatibility.  They may of course override
    other methods as desired.

    The class is missing the following attributes and methods, which don't
    really make sense for anything but real files:

        * fileno()
        * isatty()
        * encoding
        * mode
        * name
        * newlines

    Unlike standard file objects, all read methods share the same buffer
    and so can be freely mixed (e.g. read(), readline(), next(), ...).

    This class understands and will accept the following mode strings,
    with any additional characters being ignored:

        * r    - open the file for reading only.
        * r+   - open the file for reading and writing.
        * r-   - open the file for streamed reading; do not allow seek/tell.
        * w    - open the file for writing only; create the file if
                 it doesn't exist; truncate it to zero length.
        * w+   - open the file for reading and writing; create the file
                 if it doesn't exist; truncate it to zero length.
        * w-   - open the file for streamed writing; do not allow seek/tell.
        * a    - open the file for writing only; create the file if it
                 doesn't exist; place pointer at end of file.
        * a+   - open the file for reading and writing; create the file
                 if it doesn't exist; place pointer at end of file.

    These are mostly standard except for the "-" indicator, which has
    been added for efficiency purposes in cases where seeking can be
    expensive to simulate (e.g. compressed files).  Note that any file
    opened for both reading and writing must also support seeking.
    """

    def __init__(self,bufsize=1024*64):
        """FileLikeBase Constructor.

        The optional argument 'bufsize' specifies the number of bytes to
        read at a time when looking for a newline character.  Setting this to
        a larger number when lines are long should improve efficiency.
        """
        super(FileLikeBase, self).__init__()
        # File-like attributes
        self.closed = False
        self.softspace = 0
        # Our own attributes
        self._bufsize = bufsize  # buffer size for chunked reading
        self._rbuffer = None     # data that's been read but not returned
        self._wbuffer = None     # data that's been given but not written
        self._sbuffer = None     # data between real & apparent file pos
        self._soffset = 0        # internal offset of file pointer

    #
    #  The following five methods are the ones that subclasses are expected
    #  to implement.  Carefully check their docstrings.
    #

    def _read(self,sizehint=-1):
        """Read approximately <sizehint> bytes from the file-like object.

        This method is to be implemented by subclasses that wish to be
        readable.  It should read approximately <sizehint> bytes from the
        file and return them as a string.  If <sizehint> is missing or
        less than or equal to zero, try to read all the remaining contents.

        The method need not guarantee any particular number of bytes -
        it may return more bytes than requested, or fewer.  If needed the
        size hint may be completely ignored.  It may even return an empty
        string if no data is yet available.

        Because of this, the method must return None to signify that EOF
        has been reached.  The higher-level methods will never indicate EOF
        until None has been read from _read().  Once EOF is reached, it
        should be safe to call _read() again, immediately returning None.
        """
        raise NotReadableError("Object not readable")

    def _write(self,string,flushing=False):
        """Write the given string to the file-like object.

        This method must be implemented by subclasses wishing to be writable.
        It must attempt to write as much of the given data as possible to the
        file, but need not guarantee that it is all written.  It may return
        None to indicate that all data was written, or return as a string any
        data that could not be written.

        If the keyword argument 'flushing' is true, it indicates that the
        internal write buffers are being flushed, and *all* the given data
        is expected to be written to the file. If unwritten data is returned
        when 'flushing' is true, an IOError will be raised.
        """
        raise NotWritableError("Object not writable")

    def _seek(self,offset,whence):
        """Set the file's internal position pointer, approximately.

        This method should set the file's position to approximately 'offset'
        bytes relative to the position specified by 'whence'.  If it is
        not possible to position the pointer exactly at the given offset,
        it should be positioned at a convenient *smaller* offset and the
        file data between the real and apparent position should be returned.

        At minimum, this method must implement the ability to seek to
        the start of the file, i.e. offset=0 and whence=0.  If more
        complex seeks are difficult to implement then it may raise
        NotImplementedError to have them simulated (inefficiently) by
        the higher-level machinery of this class.
        """
        raise NotSeekableError("Object not seekable")

    def _tell(self):
        """Get the location of the file's internal position pointer.

        This method must be implemented by subclasses that wish to be
        seekable, and must return the position of the file's internal
        pointer.

        Due to buffering, the position seen by users of this class
        (the "apparent position") may be different to the position
        returned by this method (the "actual position").
        """
        raise NotSeekableError("Object not seekable")

    def _truncate(self,size):
        """Truncate the file's size to <size>.

        This method must be implemented by subclasses that wish to be
        truncatable.  It must truncate the file to exactly the given size
        or fail with an IOError.

        Note that <size> will never be None; if it was not specified by the
        user then it is calculated as the file's apparent position (which may
        be different to its actual position due to buffering).
        """
        raise NotTruncatableError("Object not truncatable")

    #
    #  The following methods provide the public API of the filelike object.
    #  Subclasses shouldn't need to mess with these (except perhaps for
    #  close() and flush())
    #

    def _check_mode(self,mode,mstr=None):
        """Check whether the file may be accessed in the given mode.

        'mode' must be one of "r" or "w", and this function returns False
        if the file-like object has a 'mode' attribute, and it does not
        permit access in that mode.  If there is no 'mode' attribute,
        it defaults to "r+".

        If seek support is not required, use "r-" or "w-" as the mode string.

        To check a mode string other than self.mode, pass it in as the
        second argument.
        """
        if mstr is None:
            try:
                mstr = self.mode
            except AttributeError:
                mstr = "r+"
        if "+" in mstr:
            return True
        if "-" in mstr and "-" not in mode:
            return False
        if "r" in mode:
            if "r" not in mstr:
                return False
        if "w" in mode:
            if "w" not in mstr and "a" not in mstr:
                return False
        return True

    def _assert_mode(self,mode,mstr=None):
        """Check whether the file may be accessed in the given mode.

        This method is equivalent to _check_assert(), but raises IOError
        instead of returning False.
        """
        if mstr is None:
            try:
                mstr = self.mode
            except AttributeError:
                mstr = "r+"
        if "+" in mstr:
            return True
        if "-" in mstr and "-" not in mode:
            raise NotSeekableError("File does not support seeking.")
        if "r" in mode:
            if "r" not in mstr:
                raise NotReadableError("File not opened for reading")
        if "w" in mode:
            if "w" not in mstr and "a" not in mstr:
                raise NotWritableError("File not opened for writing")
        return True

    def flush(self):
        """Flush internal write buffer, if necessary."""
        if self.closed:
            raise IOError("File has been closed")
        if self._check_mode("w-") and self._wbuffer is not None:
            buffered = b("")
            if self._sbuffer:
                buffered = buffered + self._sbuffer
                self._sbuffer = None
            buffered = buffered + self._wbuffer
            self._wbuffer = None
            leftover = self._write(buffered,flushing=True)
            if leftover and not isinstance(leftover, int):
                raise IOError("Could not flush write buffer.")

    def close(self):
        """Flush write buffers and close the file.

        The file may not be accessed further once it is closed.
        """
        #  Errors in subclass constructors can cause this to be called without
        #  having called FileLikeBase.__init__().  Since we need the attrs it
        #  initializes in cleanup, ensure we call it here.
        if not hasattr(self,"closed"):
            FileLikeBase.__init__(self)
        if not self.closed:
            self.flush()
            self.closed = True

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self,exc_type,exc_val,exc_tb):
        self.close()
        return False

    def next(self):
        """next() method complying with the iterator protocol.

        File-like objects are their own iterators, with each call to
        next() returning subsequent lines from the file.
        """
        ln = self.readline()
        if ln == b(""):
            raise StopIteration()
        return ln

    def __iter__(self):
        return self

    def truncate(self,size=None):
        """Truncate the file to the given size.

        If <size> is not specified or is None, the current file position is
        used.  Note that this method may fail at runtime if the underlying
        filelike object is not truncatable.
        """
        if "-" in getattr(self,"mode",""):
            raise NotTruncatableError("File is not seekable, can't truncate.")
        if self._wbuffer:
            self.flush()
        if size is None:
            size = self.tell()
        self._truncate(size)

    def seek(self,offset,whence=0):
        """Move the internal file pointer to the given location."""
        if whence > 2 or whence < 0:
            raise ValueError("Invalid value for 'whence': " + str(whence))
        if "-" in getattr(self,"mode",""):
            raise NotSeekableError("File is not seekable.")
        # Ensure that there's nothing left in the write buffer
        if self._wbuffer:
            self.flush()
        # Adjust for any data left in the read buffer
        if whence == 1 and self._rbuffer:
            offset = offset - len(self._rbuffer)
        self._rbuffer = None
        # Adjust for any discrepancy in actual vs apparent seek position
        if whence == 1:
            if self._sbuffer:
                offset = offset + len(self._sbuffer)
            if self._soffset:
                offset = offset + self._soffset
        self._sbuffer = None
        self._soffset = 0
        # Shortcut the special case of staying put.
        # As per posix, this has already cases the buffers to be flushed.
        if offset == 0 and whence == 1:
            return
        # Catch any failed attempts to read while simulating seek
        try:
            # Try to do a whence-wise seek if it is implemented.
            sbuf = None
            try:
                sbuf = self._seek(offset,whence)
            except NotImplementedError:
                # Try to simulate using an absolute seek.
                try:
                    if whence == 1:
                        offset = self._tell() + offset
                    elif whence == 2:
                        if hasattr(self,"size"):
                            offset = self.size + offset
                        else:
                            self._do_read_rest()
                            offset = self.tell() + offset
                    else:
                        # absolute seek already failed, don't try again
                        raise NotImplementedError
                    sbuf = self._seek(offset,0)
                except NotImplementedError:
                    # Simulate by reseting to start
                    self._seek(0,0)
                    self._soffset = offset
            finally:
                self._sbuffer = sbuf
        except NotReadableError:
            raise NotSeekableError("File not readable, can't simulate seek")

    def tell(self):
        """Determine current position of internal file pointer."""
        # Need to adjust for unread/unwritten data in buffers
        pos = self._tell()
        if self._rbuffer:
            pos = pos - len(self._rbuffer)
        if self._wbuffer:
            pos = pos + len(self._wbuffer)
        if self._sbuffer:
            pos = pos + len(self._sbuffer)
        if self._soffset:
            pos = pos + self._soffset
        return pos

    def read(self,size=-1):
        """Read at most 'size' bytes from the file.

        Bytes are returned as a string.  If 'size' is negative, zero or
        missing, the remainder of the file is read.  If EOF is encountered
        immediately, the empty string is returned.
        """
        if self.closed:
            raise IOError("File has been closed")
        self._assert_mode("r-")
        return self._do_read(size)

    def _do_read(self,size):
        """Private method to read from the file.

        This method behaves the same as self.read(), but skips some
        permission and sanity checks.  It is intended for use in simulating
        seek(), where we may want to read (and discard) information from
        a file not opened in read mode.

        Note that this may still fail if the file object actually can't
        be read from - it just won't check whether the mode string gives
        permission.
        """
        # If we were previously writing, ensure position is correct
        if self._wbuffer is not None:
            self.seek(0,1)
        # Discard any data that should have been seeked over
        if self._sbuffer:
            s = len(self._sbuffer)
            self._sbuffer = None
            self.read(s)
        elif self._soffset:
            s = self._soffset
            self._soffset = 0
            while s > self._bufsize:
                self._do_read(self._bufsize)
                s -= self._bufsize
            self._do_read(s)
        # Should the entire file be read?
        if size < 0:
            if self._rbuffer:
                data = [self._rbuffer]
            else:
                data = []
            self._rbuffer = b("")
            newData = self._read()
            while newData is not None:
                data.append(newData)
                newData = self._read()
            output = b("").join(data)
        # Otherwise, we need to return a specific amount of data
        else:
            if self._rbuffer:
                newData = self._rbuffer
                data = [newData]
            else:
                newData = b("")
                data = []
            sizeSoFar = len(newData)
            while sizeSoFar < size:
                newData = self._read(size-sizeSoFar)
                if not newData:
                    break
                data.append(newData)
                sizeSoFar += len(newData)
            data = b("").join(data)
            if sizeSoFar > size:
                # read too many bytes, store in the buffer
                self._rbuffer = data[size:]
                data = data[:size]
            else:
                self._rbuffer = b("")
            output = data
        return output

    def _do_read_rest(self):
        """Private method to read the file through to EOF."""
        data = self._do_read(self._bufsize)
        while data != b(""):
            data = self._do_read(self._bufsize)

    def readline(self,size=-1):
        """Read a line from the file, or at most <size> bytes."""
        bits = []
        indx = -1
        sizeSoFar = 0
        while indx == -1:
            nextBit = self.read(self._bufsize)
            bits.append(nextBit)
            sizeSoFar += len(nextBit)
            if not nextBit:
                break
            if size > 0 and sizeSoFar >= size:
                break
            indx = nextBit.find(b("\n"))
        # If not found, return whole string up to <size> length
        # Any leftovers are pushed onto front of buffer
        if indx == -1:
            data = b("").join(bits)
            if size > 0 and sizeSoFar > size:
                extra = data[size:]
                data = data[:size]
                self._rbuffer = extra + self._rbuffer
            return data
        # If found, push leftovers onto front of buffer
        # Add one to preserve the newline in the return value
        indx += 1
        extra = bits[-1][indx:]
        bits[-1] = bits[-1][:indx]
        self._rbuffer = extra + self._rbuffer
        return b("").join(bits)

    def readlines(self,sizehint=-1):
        """Return a list of all lines in the file."""
        return [ln for ln in self]

    def xreadlines(self):
        """Iterator over lines in the file - equivalent to iter(self)."""
        return iter(self)

    def write(self,string):
        """Write the given string to the file."""
        if self.closed:
            raise IOError("File has been closed")
        self._assert_mode("w-")
        # If we were previously reading, ensure position is correct
        if self._rbuffer is not None:
            self.seek(0, 1)
        # If we're actually behind the apparent position, we must also
        # write the data in the gap.
        if self._sbuffer:
            string = self._sbuffer + string
            self._sbuffer = None
        elif self._soffset:
            s = self._soffset
            self._soffset = 0
            try:
                string = self._do_read(s) + string
            except NotReadableError:
                raise NotSeekableError("File not readable, could not complete simulation of seek")
            self.seek(0, 0)
        if self._wbuffer:
            string = self._wbuffer + string
        leftover = self._write(string)
        if leftover is None or isinstance(leftover, int):
            self._wbuffer = b("")
            return len(string) - (leftover or 0)
        else:
            self._wbuffer = leftover
            return len(string) - len(leftover)

    def writelines(self,seq):
        """Write a sequence of lines to the file."""
        for ln in seq:
            self.write(ln)


class FileWrapper(FileLikeBase):
    """Base class for objects that wrap a file-like object.

    This class provides basic functionality for implementing file-like
    objects that wrap another file-like object to alter its functionality
    in some way.  It takes care of house-keeping duties such as flushing
    and closing the wrapped file.

    Access to the wrapped file is given by the attribute wrapped_file.
    By convention, the subclass's constructor should accept this as its
    first argument and pass it to its superclass's constructor in the
    same position.

    This class provides a basic implementation of _read() and _write()
    which just calls read() and write() on the wrapped object.  Subclasses
    will probably want to override these.
    """

    _append_requires_overwrite = False

    def __init__(self,wrapped_file,mode=None):
        """FileWrapper constructor.

        'wrapped_file' must be a file-like object, which is to be wrapped
        in another file-like object to provide additional functionality.

        If given, 'mode' must be the access mode string under which
        the wrapped file is to be accessed.  If not given or None, it
        is looked up on the wrapped file if possible.  Otherwise, it
        is not set on the object.
        """
        # This is used for working around flush/close inefficiencies
        self.__closing = False
        super(FileWrapper,self).__init__()
        self.wrapped_file = wrapped_file
        if mode is None:
            self.mode = getattr(wrapped_file,"mode","r+")
        else:
            self.mode = mode
        self._validate_mode()
        # Copy useful attributes of wrapped_file
        if hasattr(wrapped_file,"name"):
            self.name = wrapped_file.name
        # Respect append-mode setting
        if "a" in self.mode:
            if self._check_mode("r"):
                self.wrapped_file.seek(0)
            self.seek(0,2)

    def _validate_mode(self):
        """Check that various file-mode conditions are satisfied."""
        #  If append mode requires overwriting the underlying file,
        #  if must not be opened in append mode.
        if self._append_requires_overwrite:
            if self._check_mode("w"):
                if "a" in getattr(self.wrapped_file,"mode",""):
                    raise ValueError("Underlying file can't be in append mode")

    def __del__(self):
        #  Errors in subclass constructors could result in this being called
        #  without invoking FileWrapper.__init__.  Establish some simple
        #  invariants to prevent errors in this case.
        if not hasattr(self,"wrapped_file"):
            self.wrapped_file = None
        if not hasattr(self,"_FileWrapper__closing"):
            self.__closing = False
        #  Close the wrapper and the underlying file independently, so the
        #  latter is still closed on cleanup even if the former errors out.
        try:
            if FileWrapper is not None:
                super(FileWrapper,self).close()
        finally:
            if hasattr(getattr(self,"wrapped_file",None),"close"):
                self.wrapped_file.close()

    def close(self):
        """Close the object for reading/writing."""
        #  The superclass implementation of this will call flush(),
        #  which calls flush() on our wrapped object.  But we then call
        #  close() on it, which will call its flush() again!  To avoid
        #  this inefficiency, our flush() will not flush the wrapped
        #  file when we're closing.
        if not self.closed:
            self.__closing = True
            super(FileWrapper,self).close()
            if hasattr(self.wrapped_file,"close"):
                self.wrapped_file.close()

    def flush(self):
        """Flush the write buffers of the file."""
        super(FileWrapper,self).flush()
        if not self.__closing and hasattr(self.wrapped_file,"flush"):
            self.wrapped_file.flush()

    def _read(self,sizehint=-1):
        data = self.wrapped_file.read(sizehint)
        if data == b(""):
            return None
        return data

    def _write(self,string,flushing=False):
        self.wrapped_file.write(string)

    def _seek(self,offset,whence):
        self.wrapped_file.seek(offset,whence)

    def _tell(self):
        return self.wrapped_file.tell()

    def _truncate(self,size):
        return self.wrapped_file.truncate(size)


class StringIO(FileWrapper):
    """StringIO wrapper that more closely matches standard file behavior.

    This is a simple compatibility wrapper around the native StringIO class
    which fixes some corner-cases of its behavior.  Specifically:

        * adding __enter__ and __exit__ methods
        * having truncate(size) zero-fill when growing the file

    """

    def __init__(self,data=None,mode=None):
        wrapped_file = _StringIO()
        if data is not None:
            wrapped_file.write(data)
            wrapped_file.seek(0)
        super(StringIO,self).__init__(wrapped_file,mode)

    def getvalue(self):
        return self.wrapped_file.getvalue()

    def _truncate(self,size):
        pos = self.wrapped_file.tell()
        self.wrapped_file.truncate(size)
        curlen = len(self.wrapped_file.getvalue())
        if size > curlen:
            self.wrapped_file.seek(curlen)
            try:
                self.wrapped_file.write(b("\x00")*(size-curlen))
            finally:
                self.wrapped_file.seek(pos)


class SpooledTemporaryFile(FileWrapper):
    """SpooledTemporaryFile wrapper with some compatibility fixes.

    This is a simple compatibility wrapper around the native class of the
    same name, fixing some corner-cases of its behavior.  Specifically:

        * have truncate() accept a size argument
        * roll to disk is seeking past the max in-memory size
        * use improved StringIO class from this module

    """

    def __init__(self,max_size=0,mode="w+b",bufsize=-1,*args,**kwds):
        try:
            stf_args = (max_size,mode,bufsize) + args
            wrapped_file = _tempfile.SpooledTemporaryFile(*stf_args,**kwds)
            wrapped_file._file = StringIO()
            #wrapped_file._file = six.BytesIO()
            self.__is_spooled = True
        except AttributeError:
            ntf_args = (mode,bufsize) + args
            wrapped_file = _tempfile.NamedTemporaryFile(*ntf_args,**kwds)
            self.__is_spooled = False
        super(SpooledTemporaryFile,self).__init__(wrapped_file)

    def _seek(self,offset,whence):
        if self.__is_spooled:
            max_size = self.wrapped_file._max_size
            if whence == fs.SEEK_SET:
                if offset > max_size:
                    self.wrapped_file.rollover()
            elif whence == fs.SEEK_CUR:
                if offset + self.wrapped_file.tell() > max_size:
                    self.wrapped_file.rollover()
            else:
                if offset > 0:
                    self.wrapped_file.rollover()
        self.wrapped_file.seek(offset,whence)

    def _truncate(self,size):
        if self.__is_spooled:
            self.wrapped_file._file.truncate(size)
        else:
            self.wrapped_file.truncate(size)

    def fileno(self):
        return self.wrapped_file.fileno()


class LimitBytesFile(FileWrapper):
    """Filelike wrapper to limit bytes read from a stream."""

    def __init__(self,size,fileobj,*args,**kwds):
        self.size = size
        self.__remaining = size
        super(LimitBytesFile,self).__init__(fileobj,*args,**kwds)

    def _read(self,sizehint=-1):
        if self.__remaining <= 0:
            return None
        if sizehint is None or sizehint < 0 or sizehint > self.__remaining:
            sizehint = self.__remaining
        data = super(LimitBytesFile,self)._read(sizehint)
        if data is not None:
            self.__remaining -= len(data)
        return data





#------------------------------------------------------------------------------
#
#  Copyright (c) 2012, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Author: Pietro Berkes
#  Date:   Dec 2012
#
#------------------------------------------------------------------------------


from traits.api import Bool, HasTraits, Instance, Int, List
from traitsui.api import Item, TreeEditor, TreeNode, View

from traitsui.tests._tools import *


class Bogus(HasTraits):
    """ A bogus class representing a bogus tree. """

    bogus_list = List


class BogusTreeView(HasTraits):
    """ A traitsui view visualizing Bogus objects as trees. """

    bogus = Instance(Bogus)

    hide_root = Bool

    def default_traits_view(self):
        nodes = [
            TreeNode(node_for=[Bogus], children='bogus_list', label='=Bogus'),
        ]

        tree_editor = TreeEditor(
            nodes=nodes, hide_root=self.hide_root, editable=False
        )

        traits_view = View(
            Item(name='bogus', id='engine', editor=tree_editor),
            buttons=['OK'],
        )

        return traits_view


def _test_tree_editor_releases_listeners(hide_root):
    """ The TreeEditor should release the listener to the root node's children
    when it's disposed of.
    """

    with store_exceptions_on_all_threads():
        bogus = Bogus(bogus_list=[Bogus()])
        tree_editor_view = BogusTreeView(bogus=bogus, hide_root=hide_root)
        ui = tree_editor_view.edit_traits()

        # The TreeEditor sets a listener on the bogus object's children list
        notifiers_list = bogus.trait('bogus_list')._notifiers(False)
        nose.tools.assert_equal(1, len(notifiers_list))

        # Manually close the UI
        press_ok_button(ui)

        # The listener should be removed after the UI has been closed
        notifiers_list = bogus.trait('bogus_list')._notifiers(False)
        nose.tools.assert_equal(0, len(notifiers_list))


@skip_if_null
def test_tree_editor_listeners_with_shown_root():
    _test_tree_editor_releases_listeners(hide_root=False)


@skip_if_null
def test_tree_editor_listeners_with_hidden_root():
    _test_tree_editor_releases_listeners(hide_root=True)




from __future__ import absolute_import

from .toolkit import toolkit


def ColorTrait(*args, **traits):
    return toolkit().color_trait(*args, **traits)


def RGBColorTrait(*args, **traits):
    return toolkit().rgb_color_trait(*args, **traits)


def FontTrait(*args, **traits):
    return toolkit().font_trait(*args, **traits)



"""
Some functions for Python3 compatibility.

Not for general usage, the functionality in this file is exposed elsewhere

"""

import six
from six import PY3


def copy_file_to_fs(data, dst_fs, dst_path, chunk_size=64 * 1024, progress_callback=None, finished_callback=None):
    """Copy data from a string or a file-like object to a given fs/path"""
    if progress_callback is None:
        progress_callback = lambda bytes_written: None
    bytes_written = 0
    f = None
    try:
        progress_callback(bytes_written)
        if hasattr(data, "read"):
            read = data.read
            chunk = read(chunk_size)
            if isinstance(chunk, six.text_type):
                f = dst_fs.open(dst_path, 'w')
            else:
                f = dst_fs.open(dst_path, 'wb')
            write = f.write
            while chunk:
                write(chunk)
                bytes_written += len(chunk)
                progress_callback(bytes_written)
                chunk = read(chunk_size)
        else:
            if isinstance(data, six.text_type):
                f = dst_fs.open(dst_path, 'w')
            else:
                f = dst_fs.open(dst_path, 'wb')
            f.write(data)
            bytes_written += len(data)
            progress_callback(bytes_written)

        if hasattr(f, 'flush'):
            f.flush()
        if finished_callback is not None:
            finished_callback()

    finally:
        if f is not None:
            f.close()



""" Text Editor plugin definition. """


# Plugin extension-point imports.
from envisage import PluginDefinition, get_using_workbench


# Are we using the old UI plugin, or the shiny new Workbench plugin?
USING_WORKBENCH = get_using_workbench()

# Enthought plugin definition imports.
if USING_WORKBENCH:
    from envisage.workbench.action.action_plugin_definition import \
         Action, Group, Location, Menu, WorkbenchActionSet

else:
    from envisage.ui.ui_plugin_definition import \
         Action, Group, Menu, UIActions


# The plugin's globally unique identifier (also used as the prefix for all
# identifiers defined in this module).
ID = "envisage.plugins.refresh_code"


###############################################################################
# Extensions.
###############################################################################

if USING_WORKBENCH:
    refresh_code = Action(
        name          = "Refresh Code",
        description   = "Refresh application to reflect python code changes",
        accelerator   = "Ctrl+Shift+R",
        function_name = "traits.util.refresh.refresh",

        locations     = [
            Location(path="MenuBar/FileMenu/ExitGroup")
        ]
    )

    actions = WorkbenchActionSet(
        id   = ID + ".refresh_code_action_set",
        name = "Refresh Code",

        # fixme: This menus stuff should go away once we get ticket:312
        # resolved.
        #groups = [
        #    Group(
        #        id = "ToolsMenuGroup",
        #        location = Location(path="MenuBar")
        #    ),
        #    Group(
        #        id = "RefreshGroup",
        #        location = Location(path="MenuBar/ToolsMenu")
        #    ),
        #],

        #menus = [
        #    Menu(
        #        id = "ToolMenu",
        #        name = "&Tools",
        #        location = Location(path="MenuBar/ToolsMenuGroup"),
        #        groups = []
        #    ),
        #],

        actions = [refresh_code]
    )

    requires = "envisage.workbench.action"

else:
    refresh_code = Action(
        name          = "Refresh Code", # fixme: this should change
        description   = "Refresh application to reflect python code changes",
        menu_bar_path = "ToolsMenu/additions", # fixme: this should change
        accelerator   = "Ctrl+Shift+R",
        function_name = "traits.util.refresh.refresh"
    )

    actions =  UIActions(
        # fixme: This menus stuff should go away once we get ticket:312
        # resolved.
        menus = [
            Menu(
                id     = "ToolsMenu",
                name   = "&Tools",
                path   = "ToolsGroup",

                groups = [
                    Group(id = "Start"),
                    Group(id = "End"),
                ]
            ),
        ],

        actions = [refresh_code]
    )

    requires = "envisage.ui"

###############################################################################
# The plugin definition!
###############################################################################

class RefreshCodePluginDefinition(PluginDefinition):
    # The plugin's globally unique identifier.
    id = ID

    # General information about the plugin.
    name          = "Refresh Code Plugin"
    version       = "1.0.0"
    provider_name = "Enthought Inc"
    provider_url  = "www.enthought.com"
    enabled       = True

    # The Id's of the plugins that this plugin requires.
    requires = [requires]

    # The extension points offered by this plugin,
    extension_points = []

    # The contributions that this plugin makes to extension points offered by
    # either itself or other plugins.
    extensions = [actions]

#### EOF ######################################################################



#-------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   03/02/2007
#
#-------------------------------------------------------------------------

""" Defines an editor for playing animated GIF files.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

# Define the wx version dependent version of the editor:
if wx.__version__[:3] == '2.6':
    from animated_gif_editor_26 import AnimatedGIFEditor
else:
    from animated_gif_editor_28 import AnimatedGIFEditor



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   05/08/2007
#
#-------------------------------------------------------------------------

""" Traits UI editor factory for editing lists of strings.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Any, Str, Enum, List, Bool, Instance, Property

from ..basic_editor_factory import BasicEditorFactory

from ..toolkit import toolkit_object

from pyface.image_resource import ImageResource

#-------------------------------------------------------------------------
#  'ListStrEditor' editor factory class:
#-------------------------------------------------------------------------


class ListStrEditor(BasicEditorFactory):
    """ Editor factory for list of string editors.
    """

    #-- Trait Definitions ----------------------------------------------------

    # The editor class to be created:
    klass = Property

    # The optional extended name of the trait to synchronize the selection
    # values with:
    selected = Str

    # The optional extended name of the trait to synchronize the selection
    # indices with:
    selected_index = Str

    # The optional extended name of the trait to synchronize the activated value
    # with:
    activated = Str

    # The optional extended name of the trait to synchronize the activated
    # value's index with:
    activated_index = Str

    # The optional extended name of the trait to synchronize the right clicked
    # value with:
    right_clicked = Str

    # The optional extended name of the trait to synchronize the right clicked
    # value's index with:
    right_clicked_index = Str

    # Can the user edit the values?
    editable = Bool(True)

    # Are multiple selected items allowed?
    multi_select = Bool(False)

    # Should horizontal lines be drawn between items?
    horizontal_lines = Bool(False)

    # The title for the editor:
    title = Str

    # The optional extended name of the trait containing the editor title:
    title_name = Str

    # Should a new item automatically be added to the end of the list to allow
    # the user to create new entries?
    auto_add = Bool(False)

    # The adapter from list items to editor values:
    adapter = Instance('traitsui.list_str_adapter.ListStrAdapter',
                       ())

    # The optional extended name of the trait containing the adapter:
    adapter_name = Str

    # What type of operations are allowed on the list:
    operations = List(Enum('delete', 'insert', 'append', 'edit', 'move'),
                      ['delete', 'insert', 'append', 'edit', 'move'])

    # Are 'drag_move' operations allowed (i.e. True), or should they always be
    # treated as 'drag_copy' operations (i.e. False):
    drag_move = Bool(False)

    # The set of images that can be used:
    images = List(ImageResource)

    # Right-click context menu (Qt4 only). The value can be one of:
    #
    # - Instance( Menu ): Use this menu as the context menu
    # - string: Name of traits that will provide menu
    # - None: Use the default context menu
    # - False: Do not display a context menu
    menu = Any

    def _get_klass(self):
        """ Returns the editor class to be created.
        """
        return toolkit_object('list_str_editor:_ListStrEditor')
##EOF #########################################################################



"""
fs.expose.fuse
==============

Expose an FS object to the native filesystem via FUSE

This module provides the necessary interfaces to mount an FS object into
the local filesystem via FUSE::

    http://fuse.sourceforge.net/

For simple usage, the function 'mount' takes an FS object and a local path,
and exposes the given FS at that path::

    >>> from fs.memoryfs import MemoryFS
    >>> from fs.expose import fuse
    >>> fs = MemoryFS()
    >>> mp = fuse.mount(fs,"/mnt/my-memory-fs")
    >>> mp.path
    '/mnt/my-memory-fs'
    >>> mp.unmount()

The above spawns a new background process to manage the FUSE event loop, which
can be controlled through the returned subprocess.Popen object.  To avoid
spawning a new process, set the 'foreground' option::

    >>> #  This will block until the filesystem is unmounted
    >>> fuse.mount(fs,"/mnt/my-memory-fs",foreground=True)

Any additional options for the FUSE process can be passed as keyword arguments
to the 'mount' function.

If you require finer control over the creation of the FUSE process, you can
instantiate the MountProcess class directly.  It accepts all options available
to subprocess.Popen::

    >>> from subprocess import PIPE
    >>> mp = fuse.MountProcess(fs,"/mnt/my-memory-fs",stderr=PIPE)
    >>> fuse_errors = mp.communicate()[1]

The binding to FUSE is created via ctypes, using a custom version of the
fuse.py code from Giorgos Verigakis:

    https://github.com/terencehonles/fusepy

"""

import sys
if sys.platform == "win32":
    raise ImportError("FUSE is not available on win32")

import datetime
import os
import signal
import errno
import time
import stat as statinfo
import subprocess
import cPickle

import logging
logger = logging.getLogger("fs.expose.fuse")

from fs.base import flags_to_mode, threading
from fs.errors import *
from fs.path import *
from fs.local_functools import wraps

from six import PY3
from six import b

try:
    if PY3:
        from fs.expose.fuse import fuse_ctypes as fuse
    else:
        from fs.expose.fuse import fuse3 as fuse

except NotImplementedError:
    raise ImportError("FUSE found but not usable")
try:
    fuse._libfuse.fuse_get_context
except AttributeError:
    raise ImportError("could not locate FUSE library")


FUSE = fuse.FUSE
Operations = fuse.Operations
fuse_get_context = fuse.fuse_get_context

STARTUP_TIME = time.time()
NATIVE_ENCODING = sys.getfilesystemencoding()


def handle_fs_errors(func):
    """Method decorator to report FS errors in the appropriate way.

    This decorator catches all FS errors and translates them into an
    equivalent OSError.  It also makes the function return zero instead
    of None as an indication of successful execution.
    """
    name = func.__name__
    func = convert_fs_errors(func)
    @wraps(func)
    def wrapper(*args,**kwds):
        #logger.debug("CALL %r %s",name,repr(args))
        try:
            res = func(*args,**kwds)
        except Exception:
            #logger.exception("ERR %r %s",name,repr(args))
            raise
        else:
            #logger.exception("OK %r %s %r",name,repr(args),res)
            if res is None:
                return 0
        return res
    return wrapper


class FSOperations(Operations):
    """FUSE Operations interface delegating all activities to an FS object."""

    def __init__(self, fs, on_init=None, on_destroy=None):
        self.fs = fs
        self._on_init = on_init
        self._on_destroy = on_destroy
        self._files_by_handle = {}
        self._files_lock = threading.Lock()
        self._next_handle = 1
        #  FUSE expects a succesful write() to be reflected in the file's
        #  reported size, but the FS might buffer writes and prevent this.
        #  We explicitly keep track of the size FUSE expects a file to be.
        #  This dict is indexed by path, then file handle.
        self._files_size_written = {}

    def _get_file(self, fh):
        try:
            return self._files_by_handle[fh.fh]
        except KeyError:
            raise FSError("invalid file handle")

    def _reg_file(self, f, path):
        self._files_lock.acquire()
        try:
            fh = self._next_handle
            self._next_handle += 1
            lock = threading.Lock()
            self._files_by_handle[fh] = (f,path,lock)
            if path not in self._files_size_written:
                self._files_size_written[path] = {}
            self._files_size_written[path][fh] = 0
            return fh
        finally:
            self._files_lock.release()

    def _del_file(self, fh):
        self._files_lock.acquire()
        try:
            (f,path,lock) = self._files_by_handle.pop(fh.fh)
            del self._files_size_written[path][fh.fh]
            if not self._files_size_written[path]:
                del self._files_size_written[path]
        finally:
            self._files_lock.release()

    def init(self, conn):
        if self._on_init:
            self._on_init()

    def destroy(self, data):
        if self._on_destroy:
            self._on_destroy()

    @handle_fs_errors
    def chmod(self, path, mode):
        raise UnsupportedError("chmod")

    @handle_fs_errors
    def chown(self, path, uid, gid):
        raise UnsupportedError("chown")

    @handle_fs_errors
    def create(self, path, mode, fi):
        path = path.decode(NATIVE_ENCODING)
        # FUSE doesn't seem to pass correct mode information here - at least,
        # I haven't figured out how to distinguish between "w" and "w+".
        # Go with the most permissive option.
        mode = flags_to_mode(fi.flags)
        fh = self._reg_file(self.fs.open(path, mode), path)
        fi.fh = fh
        fi.keep_cache = 0

    @handle_fs_errors
    def flush(self, path, fh):
        (file, _, lock) = self._get_file(fh)
        lock.acquire()
        try:
            file.flush()
        finally:
            lock.release()

    @handle_fs_errors
    def getattr(self, path, fh=None):
        attrs = self._get_stat_dict(path.decode(NATIVE_ENCODING))
        return attrs

    @handle_fs_errors
    def getxattr(self, path, name, position=0):
        path = path.decode(NATIVE_ENCODING)
        name = name.decode(NATIVE_ENCODING)
        try:
            value = self.fs.getxattr(path, name)
        except AttributeError:
            raise UnsupportedError("getxattr")
        else:
            if value is None:
                raise OSError(errno.ENODATA, "no attribute '%s'" % (name,))
            return value

    @handle_fs_errors
    def link(self, target, souce):
        raise UnsupportedError("link")

    @handle_fs_errors
    def listxattr(self, path):
        path = path.decode(NATIVE_ENCODING)
        try:
            return self.fs.listxattrs(path)
        except AttributeError:
            raise UnsupportedError("listxattrs")

    @handle_fs_errors
    def mkdir(self, path, mode):
        path = path.decode(NATIVE_ENCODING)
        try:
            self.fs.makedir(path, recursive=True)
        except TypeError:
            self.fs.makedir(path)

    @handle_fs_errors
    def mknod(self, path, mode, dev):
        raise UnsupportedError("mknod")

    @handle_fs_errors
    def open(self, path, fi):
        path = path.decode(NATIVE_ENCODING)
        mode = flags_to_mode(fi.flags)
        fi.fh = self._reg_file(self.fs.open(path, mode), path)
        fi.keep_cache = 0
        return 0

    @handle_fs_errors
    def read(self, path, size, offset, fh):
        (file, _, lock) = self._get_file(fh)
        lock.acquire()
        try:
            file.seek(offset)
            data = file.read(size)
            return data
        finally:
            lock.release()

    @handle_fs_errors
    def readdir(self, path, fh=None):
        path = path.decode(NATIVE_ENCODING)
        entries = ['.', '..']
        for (nm, info) in self.fs.listdirinfo(path):
            self._fill_stat_dict(pathjoin(path, nm), info)
            entries.append((nm.encode(NATIVE_ENCODING), info, 0))
        return entries

    @handle_fs_errors
    def readlink(self, path):
        raise UnsupportedError("readlink")

    @handle_fs_errors
    def release(self, path, fh):
        (file, _, lock) = self._get_file(fh)
        lock.acquire()
        try:
            file.close()
            self._del_file(fh)
        finally:
            lock.release()

    @handle_fs_errors
    def removexattr(self, path, name):
        path = path.decode(NATIVE_ENCODING)
        name = name.decode(NATIVE_ENCODING)
        try:
            return self.fs.delxattr(path, name)
        except AttributeError:
            raise UnsupportedError("removexattr")

    @handle_fs_errors
    def rename(self, old, new):
        old = old.decode(NATIVE_ENCODING)
        new = new.decode(NATIVE_ENCODING)
        try:
            self.fs.rename(old, new)
        except FSError:
            if self.fs.isdir(old):
                self.fs.movedir(old, new)
            else:
                self.fs.move(old, new)

    @handle_fs_errors
    def rmdir(self, path):
        path = path.decode(NATIVE_ENCODING)
        self.fs.removedir(path)

    @handle_fs_errors
    def setxattr(self, path, name, value, options, position=0):
        path = path.decode(NATIVE_ENCODING)
        name = name.decode(NATIVE_ENCODING)
        try:
            return self.fs.setxattr(path, name, value)
        except AttributeError:
            raise UnsupportedError("setxattr")

    @handle_fs_errors
    def symlink(self, target, source):
        raise UnsupportedError("symlink")

    @handle_fs_errors
    def truncate(self, path, length, fh=None):
        path = path.decode(NATIVE_ENCODING)
        if fh is None and length == 0:
            self.fs.open(path, "wb").close()
        else:
            if fh is None:
                f = self.fs.open(path, "rb+")
                if not hasattr(f, "truncate"):
                    raise UnsupportedError("truncate")
                f.truncate(length)
            else:
                (file, _, lock) = self._get_file(fh)
                lock.acquire()
                try:
                    if not hasattr(file, "truncate"):
                        raise UnsupportedError("truncate")
                    file.truncate(length)
                finally:
                    lock.release()
        self._files_lock.acquire()
        try:
            try:
                size_written = self._files_size_written[path]
            except KeyError:
                pass
            else:
                for k in size_written:
                    size_written[k] = length
        finally:
            self._files_lock.release()

    @handle_fs_errors
    def unlink(self, path):
        path = path.decode(NATIVE_ENCODING)
        self.fs.remove(path)

    @handle_fs_errors
    def utimens(self, path, times=None):
        path = path.decode(NATIVE_ENCODING)
        accessed_time, modified_time = times
        if accessed_time is not None:
            accessed_time = datetime.datetime.fromtimestamp(accessed_time)
        if modified_time is not None:
            modified_time = datetime.datetime.fromtimestamp(modified_time)
        self.fs.settimes(path, accessed_time, modified_time)

    @handle_fs_errors
    def write(self, path, data, offset, fh):
        (file, path, lock) = self._get_file(fh)
        lock.acquire()
        try:
            file.seek(offset)
            file.write(data)
            if self._files_size_written[path][fh.fh] < offset + len(data):
                self._files_size_written[path][fh.fh] = offset + len(data)
            return len(data)
        finally:
            lock.release()

    def _get_stat_dict(self, path):
        """Build a 'stat' dictionary for the given file."""
        info = self.fs.getinfo(path)
        self._fill_stat_dict(path, info)
        return info

    def _fill_stat_dict(self, path, info):
        """Fill default values in the stat dict."""
        uid, gid, pid = fuse_get_context()
        private_keys = [k for k in info if k.startswith("_")]
        for k in private_keys:
            del info[k]
        #  Basic stuff that is constant for all paths
        info.setdefault("st_ino", 0)
        info.setdefault("st_dev", 0)
        info.setdefault("st_uid", uid)
        info.setdefault("st_gid", gid)
        info.setdefault("st_rdev", 0)
        info.setdefault("st_blksize", 1024)
        info.setdefault("st_blocks", 1)
        #  The interesting stuff
        if 'st_mode' not in info:
            if self.fs.isdir(path):
                info['st_mode'] = 0755
            else:
                info['st_mode'] = 0666
        mode = info['st_mode']
        if not statinfo.S_ISDIR(mode) and not statinfo.S_ISREG(mode):
            if self.fs.isdir(path):
                info["st_mode"] = mode | statinfo.S_IFDIR
                info.setdefault("st_nlink", 2)
            else:
                info["st_mode"] = mode | statinfo.S_IFREG
                info.setdefault("st_nlink", 1)
        for (key1, key2) in [("st_atime", "accessed_time"), ("st_mtime", "modified_time"), ("st_ctime", "created_time")]:
            if key1 not in info:
                if key2 in info:
                    info[key1] = time.mktime(info[key2].timetuple())
                else:
                    info[key1] = STARTUP_TIME
        #  Ensure the reported size reflects any writes performed, even if
        #  they haven't been flushed to the filesystem yet.
        if "size" in info:
            info["st_size"] = info["size"]
        elif "st_size" not in info:
            info["st_size"] = 0
        try:
            written_sizes = self._files_size_written[path]
        except KeyError:
            pass
        else:
            info["st_size"] = max(written_sizes.values() + [info["st_size"]])
        return info


def mount(fs, path, foreground=False, ready_callback=None, unmount_callback=None, **kwds):
    """Mount the given FS at the given path, using FUSE.

    By default, this function spawns a new background process to manage the
    FUSE event loop.  The return value in this case is an instance of the
    'MountProcess' class, a subprocess.Popen subclass.

    If the keyword argument 'foreground' is given, we instead run the FUSE
    main loop in the current process.  In this case the function will block
    until the filesystem is unmounted, then return None.

    If the keyword argument 'ready_callback' is provided, it will be called
    when the filesystem has been mounted and is ready for use.  Any additional
    keyword arguments will be passed through as options to the underlying
    FUSE class.  Some interesting options include:

        * nothreads Switch off threading in the FUSE event loop
        * fsname Name to display in the mount info table

    """
    path = os.path.expanduser(path)
    if foreground:
        op = FSOperations(fs, on_init=ready_callback, on_destroy=unmount_callback)
        return FUSE(op, path, raw_fi=True, foreground=foreground, **kwds)
    else:
        mp = MountProcess(fs, path, kwds)
        if ready_callback:
            ready_callback()
        if unmount_callback:
            orig_unmount = mp.unmount

            def new_unmount():
                orig_unmount()
                unmount_callback()
            mp.unmount = new_unmount
        return mp


def unmount(path):
    """Unmount the given mount point.

    This function shells out to the 'fusermount' program to unmount a
    FUSE filesystem.  It works, but it would probably be better to use the
    'unmount' method on the MountProcess class if you have it.

    On darwin, "diskutil umount <path>" is called
    On freebsd, "umount <path>" is called
    """
    if sys.platform == "darwin":
        args = ["diskutil", "umount", path]
    elif "freebsd" in sys.platform:
        args = ["umount", path]
    else:
        args = ["fusermount", "-u", path]

    for num_tries in xrange(3):
        p = subprocess.Popen(args,
                             stderr=subprocess.PIPE,
                             stdout=subprocess.PIPE)
        (stdout, stderr) = p.communicate()
        if p.returncode == 0:
            return
        if "not mounted" in stderr:
            return
        if "not found" in stderr:
            return
    raise OSError("filesystem could not be unmounted: %s (%s) " % (path, str(stderr).rstrip(),))


class MountProcess(subprocess.Popen):
    """subprocess.Popen subclass managing a FUSE mount.

    This is a subclass of subprocess.Popen, designed for easy management of
    a FUSE mount in a background process.  Rather than specifying the command
    to execute, pass in the FS object to be mounted, the target mount point
    and a dictionary of options for the underlying FUSE class.

    In order to be passed successfully to the new process, the FS object
    must be pickleable. This restriction may be lifted in the future.

    This class has an extra attribute 'path' giving the path to the mounted
    filesystem, and an extra method 'unmount' that will cleanly unmount it
    and terminate the process.

    By default, the spawning process will block until it receives notification
    that the filesystem has been mounted.  Since this notification is sent
    by writing to a pipe, using the 'close_fds' option on this class will
    prevent it from being sent.  You can also pass in the keyword argument
    'nowait' to continue without waiting for notification.

    """

    #  This works by spawning a new python interpreter and passing it the
    #  pickled (fs,path,opts) tuple on the command-line.  Something like this:
    #
    #    python -c "import MountProcess; MountProcess._do_mount('..data..')
    #
    #  It would be more efficient to do a straight os.fork() here, and would
    #  remove the need to pickle the FS.  But API wise, I think it's much
    #  better for mount() to return a Popen instance than just a pid.
    #
    #  In the future this class could implement its own forking logic and
    #  just copy the relevant bits of the Popen interface.  For now, this
    #  spawn-a-new-interpreter solution is the easiest to get up and running.

    unmount_timeout = 5

    def __init__(self, fs, path, fuse_opts={}, nowait=False, **kwds):
        self.path = path
        if nowait or kwds.get("close_fds", False):
            if PY3:
                cmd = "from pickle import loads;"
            else:
                cmd = "from cPickle import loads;"
            #cmd = 'import cPickle; '
            cmd = cmd + 'data = loads(%s); '
            cmd = cmd + 'from fs.expose.fuse import MountProcess; '
            cmd = cmd + 'MountProcess._do_mount_nowait(data)'
            cmd = cmd % (repr(cPickle.dumps((fs, path, fuse_opts), -1)),)
            cmd = [sys.executable, "-c", cmd]
            super(MountProcess, self).__init__(cmd, **kwds)
        else:
            (r, w) = os.pipe()
            if PY3:
                cmd = "from pickle import loads;"
            else:
                cmd = "from cPickle import loads;"
            #cmd = 'import cPickle; '
            cmd = cmd + 'data = loads(%s); '
            cmd = cmd + 'from fs.expose.fuse import MountProcess; '
            cmd = cmd + 'MountProcess._do_mount_wait(data)'
            cmd = cmd % (repr(cPickle.dumps((fs, path, fuse_opts, r, w), -1)),)
            cmd = [sys.executable, "-c", cmd]
            super(MountProcess, self).__init__(cmd, **kwds)
            os.close(w)

            byte = os.read(r, 1)
            if byte != b("S"):
                err_text = os.read(r, 20)
                self.terminate()
                if hasattr(err_text, 'decode'):
                    err_text = err_text.decode(NATIVE_ENCODING)
                raise RuntimeError("FUSE error: " + err_text)

    def unmount(self):
        """Cleanly unmount the FUSE filesystem, terminating this subprocess."""
        self.terminate()
        def killme():
            self.kill()
            time.sleep(0.1)
            try:
                unmount(self.path)
            except OSError:
                pass
        tmr = threading.Timer(self.unmount_timeout, killme)
        tmr.start()
        self.wait()
        tmr.cancel()

    if not hasattr(subprocess.Popen, "terminate"):
        def terminate(self):
            """Gracefully terminate the subprocess."""
            os.kill(self.pid, signal.SIGTERM)

    if not hasattr(subprocess.Popen, "kill"):
        def kill(self):
            """Forcibly terminate the subprocess."""
            os.kill(self.pid, signal.SIGKILL)

    @staticmethod
    def _do_mount_nowait(data):
        """Perform the specified mount, return without waiting."""
        fs, path, opts = data
        opts["foreground"] = True

        def unmount_callback():
            fs.close()
        opts["unmount_callback"] = unmount_callback
        mount(fs, path, *opts)

    @staticmethod
    def _do_mount_wait(data):
        """Perform the specified mount, signalling when ready."""
        fs, path, opts, r, w = data
        os.close(r)
        opts["foreground"] = True
        successful = []

        def ready_callback():
            successful.append(True)
            os.write(w, b("S"))
            os.close(w)
        opts["ready_callback"] = ready_callback

        def unmount_callback():
            fs.close()
        opts["unmount_callback"] = unmount_callback
        try:
            mount(fs, path, **opts)
        except Exception, e:
            os.write(w, b("E") + unicode(e).encode('ascii', errors='replace'))
            os.close(w)

        if not successful:
            os.write(w, b("EMount unsuccessful"))
            os.close(w)


if __name__ == "__main__":
    import os
    import os.path
    from fs.tempfs import TempFS
    mount_point = os.path.join(os.environ["HOME"], "fs.expose.fuse")
    if not os.path.exists(mount_point):
        os.makedirs(mount_point)

    def ready_callback():
        print "READY"
    mount(TempFS(), mount_point, foreground=True, ready_callback=ready_callback)
